import warnings

import requests

from ..util.decorator import check_null
from .util import DEFAULT_RES
from .util import MapKeys
from .util import MapUrls
from .util import error_baidu
from .util import fix_address
from .util import fix_city
from .util import gcj2xx
from .util import rv_score

KEY = MapKeys.baidu


@check_null()
def address_json(address: str, city: str, key=None):
  """百度地理编码接口"""
  if not key:
    key = KEY
  url = f'{MapUrls.baidu}?city={city}&address={address}&output=json&ak={key}&ret_coordtype=gcj02ll'
  js = requests.get(url).json()
  error_baidu(js)
  if js['status'] == 0:
    return js['result']


@check_null()
def place_json(address: str, city: str, key=None):
  """百度地点检索接口"""
  if not key:
    key = KEY
  url = f'{MapUrls.baidu_poi}?query={address}&region={city}&city_limit=true&output=json&ak={key}&ret_coordtype=gcj02ll'
  js = requests.get(url).json()
  error_baidu(js)
  if js['status'] == 0 and len(js['results']) >= 1:
    return js['results'][0]


@check_null()
def get_baidu(*,
              address,
              city,
              source,
              disable_cache=False,
              with_detail=True,
              key=None):
  """MC geocode服务"""
  assert source in ('baidu', 'baidu_poi')
  url = f'{MapUrls.mdt}?address={address}&city={city}&disable_cache={disable_cache}&with_detail={with_detail}&source={source}'
  req = requests.get(url)
  if req.status_code == 200:
    return req.json()['result'][0]['extra']
  if req.status_code in (400, 403):
    if source == 'baidu':
      return address_json(address=address, city=city, key=key)
    if source == 'baidu_poi':
      return place_json(address=address, city=city, key=key)
  else:
    warnings.warn(f'Unexpected status_code：{req.status_code}，{city}|{address}')


@check_null(default_rv=DEFAULT_RES)
def get_address_baidu(address: str, city: str, srs='wgs84', key=None) -> dict:
  result = DEFAULT_RES.copy()
  source = 'baidu'
  city = fix_city(city)
  address = fix_address(address)
  address_dict = get_baidu(address=address, city=city, source=source, key=key)
  if address_dict:
    if address_dict['precise'] == 1:
      result['score'] = 100
    elif _score := address_dict.get('comprehension'):
      result['score'] = _score
    else:
      result['score'] = address_dict.get('confidence')
    latlng = gcj2xx(
        [address_dict['location']['lng'], address_dict['location']['lat']],
        srs=srs
    )
    result['lng'] = latlng[1]
    result['lat'] = latlng[0]
    result['source'] = source
  return result


@check_null(default_rv=DEFAULT_RES)
def get_place_baidu(address: str, city: str, srs='wgs84', key=None) -> dict:
  result = DEFAULT_RES.copy()
  source = 'baidu_poi'
  city = fix_city(city)
  address = fix_address(address)
  poi_dict = get_baidu(address=address, city=city, source=source, key=key)
  if poi_dict:
    result['rv'] = poi_dict['name']
    latlng = gcj2xx(
        [poi_dict['location']['lng'], poi_dict['location']['lat']],
        srs=srs
    )
    result['lng'] = latlng[1]
    result['lat'] = latlng[0]
    result['score'] = rv_score(city, address, poi_dict['name'])
    result['source'] = source
  return result



import geopandas as gpd
from geopandas.testing import assert_geodataframe_equal
from shapely.geometry import Polygon

from ricco.geometry.topology_check import fix_polygon_topology


def assert_geopandas(res: gpd.GeoDataFrame, test: gpd.GeoDataFrame):
  diff1 = res.overlay(test, how='difference', keep_geom_type=False)
  diff2 = test.overlay(res, how='difference', keep_geom_type=False)
  assert (diff1.empty or all(diff1['geometry'].area < 1e-5))
  assert (diff2.empty or all(diff2['geometry'].area < 1e-5))
  assert_geodataframe_equal(res[res['geometry'].is_empty],
                            test[test['geometry'].is_empty],
                            check_crs=False)


def test_fix_overlap():
  polygon1 = Polygon([(0, 0), (0, 4), (4, 4), (4, 0)])
  polygon2 = Polygon([(3, 0), (3, 4), (7, 4), (7, 0)])
  polygon4 = Polygon([(1, 1), (1, 3), (3, 3), (3, 1)])
  input1 = gpd.GeoDataFrame({'name': [0, 1],
                             'geometry': [polygon4, polygon1]},
                            geometry='geometry')
  res1 = fix_polygon_topology(input1, fill_intersects=True, keep_contains=False)
  res1.crs = None
  test1 = gpd.GeoDataFrame({'name': [0, 1],
                            'geometry': [
                              Polygon(),
                              polygon1
                            ]})
  assert_geopandas(res1, test1)

  input2 = gpd.GeoDataFrame({'name': [0, 1],
                             'geometry': [polygon4, polygon1]},
                            geometry='geometry')
  res2 = fix_polygon_topology(input2, fill_intersects=True, keep_contains=True)
  res2.crs = None

  test2 = gpd.GeoDataFrame({'name': [0, 1],
                            'geometry': [
                              polygon4,
                              polygon1.difference(polygon4.buffer(1e-7))
                            ]})
  assert_geopandas(res2, test2)

  input3 = gpd.GeoDataFrame({'name': [0, 1],
                             'geometry': [polygon2, polygon1]},
                            geometry='geometry')

  res3 = fix_polygon_topology(input3, fill_intersects=False)
  res3.crs = None

  test3 = gpd.GeoDataFrame({'name': [0, 1],
                            'geometry': [
                              polygon2.difference(polygon1.buffer(1e-7)),
                              polygon1.difference(polygon2.buffer(1e-7)),
                            ]})
  assert_geopandas(res3, test3)

  polygon5 = Polygon([(1, 1), (1, 2), (2, 2), (2, 1)])
  input4 = gpd.GeoDataFrame({
    'name': [0, 1, 2],
    'geometry': [
      polygon1,
      polygon5,
      polygon4,
    ]
  })
  res4 = fix_polygon_topology(input4, fill_intersects=True, keep_contains=True)
  res4.crs = None
  test4 = gpd.GeoDataFrame({'name': [0, 1, 2],
                            'geometry': [polygon1.difference(polygon4),
                                         polygon5,
                                         polygon4.difference(polygon5)]})
  assert_geopandas(res4, test4)



import numpy as np

from ricco.util.address_tools import is_std_hao
from ricco.util.address_tools import main
from ricco.util.address_tools import pre_process


def test_is_std_hao():
  assert is_std_hao('255号') is True
  assert is_std_hao('255号、256号') is True
  assert is_std_hao('258号、256号、256号') is True
  assert is_std_hao('255') is False
  assert is_std_hao('258、256号') is False
  assert is_std_hao('258号，256号') is False


def test_pre_process():
  assert pre_process(None) is None
  assert pre_process(np.nan) is None
  assert pre_process('255-1') == '255-1'
  assert pre_process('255—1') == '255-1'
  assert pre_process('255－1') == '255-1'
  assert pre_process('255~1') == '255-1'
  assert pre_process('255～1') == '255-1'
  assert pre_process('255～1?') == '255-1'
  assert pre_process('255～1？') == '255-1'
  assert pre_process('255～1 甲') == '255-1甲'
  assert pre_process('255乙号') == '255号乙'


def test_main():
  assert main('上海市普陀区甘泉路街道黄陵路255号') == {
    '省': None,
    '市': '上海市',
    '区': '普陀区',
    '街道': '甘泉路街道',
    '居委': None,
    '路': '黄陵路',
    '弄': None,
    '号': '255号',
    'subfix': None,
    'extra': '255号',
    'format': '黄陵路255号',
  }
  assert main('黄陵路255号-2') == {
    '省': None, '市': None, '区': None, '街道': None, '居委': None,
    '路': '黄陵路',
    '弄': None,
    '号': '255号',
    'subfix': '2',
    'extra': '255号-2',
    'format': '黄陵路255号-2',
  }
  assert main('黄陵路255') == {
    '省': None, '市': None, '区': None, '街道': None, '居委': None,
    '路': '黄陵路',
    '弄': None,
    '号': '255号',
    'subfix': None,
    'extra': '255',
    'format': '黄陵路255号',
  }
  assert main('宜君路51—1号') == {
    '省': None, '市': None, '区': None, '街道': None, '居委': None,
    '路': '宜君路',
    '弄': None,
    '号': '51号',
    'subfix': '1',
    'extra': '51-1号',
    'format': '宜君路51号-1',
  }
  assert main('宜君路51－1号') == {
    '省': None, '市': None, '区': None, '街道': None, '居委': None,
    '路': '宜君路',
    '弄': None,
    '号': '51号',
    'subfix': '1',
    'extra': '51-1号',
    'format': '宜君路51号-1',
  }
  assert main('宜君路51-1号') == {
    '省': None, '市': None, '区': None, '街道': None, '居委': None,
    '路': '宜君路',
    '弄': None,
    '号': '51号',
    'subfix': '1',
    'extra': '51-1号',
    'format': '宜君路51号-1',
  }
  assert main('宜川路351弄24号') == {
    '省': None, '市': None, '区': None, '街道': None, '居委': None,
    '路': '宜川路',
    '弄': '351弄',
    '号': '24号',
    'subfix': None,
    'extra': '24号',
    'format': '宜川路351弄24号',
  }
  assert main('宜川路351弄24号-3') == {
    '省': None, '市': None, '区': None, '街道': None, '居委': None,
    '路': '宜川路',
    '弄': '351弄',
    '号': '24号',
    'subfix': '3',
    'extra': '24号-3',
    'format': '宜川路351弄24号-3',
  }
  assert main('延长西路56号乙') == {
    '省': None, '市': None, '区': None, '街道': None, '居委': None,
    '路': '延长西路',
    '弄': None,
    '号': '56号',
    'subfix': '乙',
    'extra': '56号乙',
    'format': '延长西路56号-乙',
  }
  assert main('延长西路56甲号') == {
    '省': None, '市': None, '区': None, '街道': None, '居委': None,
    '路': '延长西路',
    '弄': None,
    '号': '56号',
    'subfix': '甲',
    'extra': '56号甲',
    'format': '延长西路56号-甲',
  }
  assert main('新村路31，33号') == {
    '省': None, '市': None, '区': None, '街道': None, '居委': None,
    '路': '新村路',
    '弄': None,
    '号': '31号、33号',
    'subfix': None,
    'extra': '31，33号',
    'format': '新村路31号、33号',
  }



import csv
import os

import pandas as pd

from ..geometry.df import auto2shapely
from ..util.exception import UnknownFileTypeError
from ..util.os import ensure_dirpath_exist
from ..util.os import extension
from .transformer import df_iter


def to_csv_by_line(data: list, filename: str):
  """
  逐行写入csv文件
  Args:
    data: 要写入的数据列表
    filename: 文件名
  """
  with open(filename, 'a') as f:
    csv_write = csv.writer(f, dialect='excel')
    csv_write.writerow(data)


def to_sheets(data: dict, filename: str, index=False):
  """
  将多个dataframe保存到不同的sheet中
  Args:
    data: 要保存的数据集，格式为：{sheet_name: DataFrame}
    filename: 要保存的文件名
    index: 是否保存index
  """
  assert isinstance(data, dict), 'data must be dict'
  with pd.ExcelWriter(filename) as writer:
    for sheet_name, _df in data.items():
      _df.to_excel(writer, sheet_name, index=index)


def to_file(df: pd.DataFrame, filepath,
            *,
            index=False,
            log=True,
            encoding=None):
  """将df保存为文件"""
  ensure_dirpath_exist(filepath)
  if log:
    print(f'Saving: {filepath}, Rows：{df.shape[0]}')
  df = df.copy()
  ex = extension(filepath)
  if ex == '.csv':
    df.to_csv(filepath, index=index, encoding=encoding)
  elif ex == '.parquet':
    df.to_parquet(filepath, index=index)
  elif ex in ('.xlsx', '.xls'):
    df.to_excel(filepath, index=index)
  elif ex in ('.shp', '.geojson'):
    df = auto2shapely(df)
    df.to_file(filepath, encoding=encoding,
               driver='GeoJSON' if ex == '.geojson' else None)
  else:
    raise UnknownFileTypeError(f'不支持的文件扩展名：{ex}')


def to_parts_file(df, dirpath,
                  chunksize=None,
                  parts=None,
                  to_ext='.csv',
                  **kwargs):
  """
  将df保存为多个文件
  Args:
    df: 要拆分保存的DataFrame
    dirpath: 文件保存的目录
    chunksize: 拆分保存的文件大小
    parts: 拆分保存的文件数量
    to_ext: 文件扩展名
  """
  for i, _df in enumerate(df_iter(df, chunksize=chunksize, parts=parts)):
    savefile = os.path.join(dirpath, f'part_{str(i).zfill(6)}{to_ext}')
    to_file(_df, savefile, **kwargs)



import pandas as pd

from ..util.assertion import assert_not_null
from ..util.base import is_empty
from ..util.decorator import run_once
from ..util.decorator import singleton


@singleton
def get_graph_dict(graph_df: (pd.DataFrame, dict),
                   c_key='id',
                   c_level_type='level_type',
                   c_parent_key='parent_id',
                   c_parent_type='parent_type'
                   ):
  graph_df = graph_df[[c_key, c_level_type, c_parent_key]]
  df_temp = graph_df[[c_key, c_level_type]]
  df_temp.columns = [c_parent_key, c_parent_type]
  graph_df = graph_df.merge(df_temp)
  # 构造查询字典
  return graph_df.set_index(c_key).to_dict()


def query_from_graph(key,
                     graph_df: (pd.DataFrame, dict),
                     c_key='id',
                     c_level_type='level_type',
                     c_parent_key='parent_id',
                     max_depth=10,
                     ) -> dict:
  """
  从图数据中查询全部父级节点
  Args:
    key: 要查询的节点
    graph_df: 要查询的数据集
    c_key: 子节点关键列的字段名
    c_level_type: 等级类型字段名
    c_parent_key: 父节点字段名
    max_depth: 最大深度
  """

  @run_once
  def asserts():
    assert isinstance(graph_df, pd.DataFrame)
    assert graph_df[c_key].is_unique
    assert_not_null(graph_df, c_level_type), f'{c_level_type}不能为空'
    assert graph_df[graph_df[c_key] == graph_df[c_parent_key]].empty, '父子不能相同'

  asserts()
  # 构造查询字典
  c_parent_type = 'parent_type'
  graph_dict = get_graph_dict(graph_df=graph_df, c_key=c_key,
                              c_level_type=c_level_type,
                              c_parent_key=c_parent_key,
                              c_parent_type=c_parent_type)
  level_type = graph_dict[c_level_type].get(key)
  if not level_type:
    return {}
  res = {level_type: key}
  # 循环查询上级关系
  depth = 0
  while depth < max_depth:
    parent_key = graph_dict[c_parent_key].get(key)
    parent_type = graph_dict[c_parent_type].get(key)
    if is_empty(parent_key):
      break
    res[parent_type] = parent_key
    key = parent_key
    depth += 1
  return res



import numpy as np
import pandas as pd

from ..util.base import ensure_list


def standard_e(df: pd.DataFrame, columns: list):
  """熵值法的正向指标的极差标准化"""
  df = df.copy()
  columns = ensure_list(columns)
  for c in columns:
    mmin = df[c].min()
    mmax = df[c].max()
    if mmin == mmax:
      df.loc[:, c] = None
    else:
      df.loc[:, c] = (df[c] - mmin) / (mmax - mmin)
  return df


def standard_e_neg(df: pd.DataFrame, columns: list):
  """熵值法的负向指标的极差标准化"""
  df = df.copy()
  columns = ensure_list(columns)
  for c in columns:
    mmin = df[c].min()
    mmax = df[c].max()
    if mmin == mmax:
      df.loc[:, c] = None
    else:
      df.loc[:, c] = (mmax - df[c]) / (mmax - mmin)
  return df


def pvalue(df: pd.DataFrame, columns: list):
  """计算指标下任一数值在该指标中的比重"""
  df = df.copy()
  columns = ensure_list(columns)
  for c in columns:
    df.loc[:, c] = df[c] / df[c].sum(axis=0)
  return df


def gevalue(df: pd.DataFrame, columns: list):
  """计算指标的差异系数"""
  df = df.copy()
  columns = ensure_list(columns)
  n = df.shape[0]
  for c in columns:
    df[c] = df.apply(lambda x: np.log(x[c]) * x[c], axis=1)
    j_sum = df[c].sum(axis=0)
    df.loc[:, c] = 1 - (-1 / np.log(n)) * j_sum
  return df


def wvalue(df, columns: list):
  """计算指标之间的权重关系"""
  df = df.copy()
  columns = ensure_list(columns)
  df_sum = df[columns].sum(axis=1)
  for c in columns:
    df.loc[:, c] = df[c] / df_sum
  return df


def entropy(df, columns: list = None):
  """
  采用熵值法计算的总分评价
  Args:
    df: 非负数化处理后的数据，默认其中数值量纲都是越大越好
    columns: 需要用熵值法计算权重的指标合集
  Returns:
    用过熵值法得到的总分, 权重
  """

  if not columns:
    columns = df.describe().columns.tolist()

  columns = ensure_list(columns)
  df = df[columns]
  df_p = pvalue(df, columns)
  df_g = gevalue(df_p, columns)
  df_w = wvalue(df_g, columns)
  x = (df_w * df_p).sum(axis=1)
  entropy_weight = df_w.median()
  return x, entropy_weight


def pca_score(df, columns: list = None):
  """PCA法计算器"""
  from sklearn.decomposition import PCA
  from sklearn.preprocessing import StandardScaler
  if not columns:
    columns = df.describe().columns.tolist()
  columns = ensure_list(columns)
  df = df[columns]
  scaler_std = StandardScaler()
  x_std = scaler_std.fit_transform(df)
  pca = PCA(n_components=1)
  x_reduced = pca.fit_transform(x_std)
  if (pca.components_[0] < 0).sum(0) == pca.components_[0].shape[0]:
    x_reduced = -x_reduced
  return x_reduced


class EntropyClass:
  def __init__(self, df, cols, neg_cols, key):
    """
    熵值法计算器-输出结果需要后续再进行rescale处理
    Args:
      df: 为原始待计算df
      cols: 需要参与计算的列名list
      neg_cols: cols中需要负向处理的列名
      key: 计算结果的列名
    Returns:
      entropy_res: 单独计算结果列
      entropy_df: 包含计算结果的df
      entropy_weight: 权重dict
    """
    df = standard_e(df, cols)
    df = standard_e_neg(df, neg_cols)
    df[key], self.entropy_weight = entropy(df, columns=cols)
    self.entropy_res = df[key]
    self.entropy_df = df


def rescale(
    df: pd.DataFrame,
    key: str,
    scale_min: float = 0,
    scale_max: float = 1,
    score_range: tuple = (0, 100)) -> pd.DataFrame:
  """
  归一化处理
  Args:
    df: 为原始待计算df
    key: 需要参与计算的列名
    scale_min: 下分位数, 低于该分位数的结果均取最小值
    scale_max: 上分位数, 高于该分位数的结果均取最大值
    score_range: 得分范围
  """
  from sklearn.preprocessing import MinMaxScaler
  scale_mm = MinMaxScaler(score_range)
  mmin = df[[key]].quantile(scale_min)[key]
  mmax = df[[key]].quantile(scale_max)[key]
  df.loc[df[key] < mmin, key] = mmin
  df.loc[df[key] > mmax, key] = mmax
  df_key = pd.DataFrame(scale_mm.fit_transform(df[[key]]),
                        index=df.index,
                        columns=[key])
  return df_key[key]



import warnings

import pandas as pd
from pandas.testing import assert_frame_equal
from shapely.geometry import MultiPolygon
from shapely.geometry import Point
from shapely.geometry import Polygon

from ricco.geometry.df import mark_tags_v2
from ricco.geometry.util import GeomFormat
from ricco.geometry.util import get_epsg
from ricco.geometry.util import infer_geom_format
from ricco.geometry.util import is_geojson
from ricco.geometry.util import is_shapely
from ricco.geometry.util import is_wkb
from ricco.geometry.util import is_wkt

point_name = '点位A'
point_lng = 121.505563
point_lat = 31.31038
point_shapely = Point(point_lng, point_lat)
point_wkb = '0101000020E610000054C4E9245B605E401D554D10754F3F40'

polygon_name = '板块B'
p1 = Point(121.4952839926596653, 31.3133866150263032)
p2 = Point(121.5195342448080282, 31.3134999823890041)
p3 = Point(121.5116388138760612, 31.2944240039077783)
polygon_shapely = MultiPolygon([Polygon([p1, p2, p3, p1])])
polygon_wkb = '0106000020E6100000010000000103000000010000000400000028ADA1BBB25F5E40C88AEE1A3A503F4090A68F0C40615E4048A8EB8841503F40903DB9B0BE605E40F8F01B5F5F4B3F4028ADA1BBB25F5E40C88AEE1A3A503F40'
polygon_wkt = 'MULTIPOLYGON (((121.4952839926596653 31.3133866150263032, 121.5195342448080282 31.3134999823890041, 121.5116388138760612 31.2944240039077783, 121.4952839926596653 31.3133866150263032)))'

point_df = pd.DataFrame({
  'name': [point_name], 'lng': [point_lng], 'lat': [point_lat],
  'geometry_shapely': [point_shapely], 'geometry_wkb': [point_wkb],
})
point_df['geometry_shapely'] = point_df['geometry_shapely'].astype('geometry')

polygon_df = pd.DataFrame({
  '板块': [polygon_name], 'geometry_wkb': [polygon_wkb],
  'geometry_shapely': [polygon_shapely],
})
polygon_df['geometry_shapely'] = polygon_df['geometry_shapely'].astype(
    'geometry')

res_df = pd.DataFrame({
  'name': [point_name], 'lng': [point_lng], 'lat': [point_lat],
  'geometry_shapely': [point_shapely], 'geometry_wkb': [point_wkb],
  '板块': [polygon_name]
})
res_df['geometry_shapely'] = res_df['geometry_shapely'].astype('geometry')


def test_get_epsg():
  assert get_epsg('郑州') == 32649
  assert get_epsg('郑州市') == 32649
  assert get_epsg('上海') == 32651
  # test warning
  with warnings.catch_warnings(record=True) as w:
    warnings.simplefilter("always")
    city = '一个没有的市'
    assert get_epsg(city) == 32649
    assert len(w) == 1
    assert issubclass(w[-1].category, UserWarning)
    assert f'请补充"{city}"的epsg信息，默认返回经度113.0' in str(w[-1].message)


def test_mark_tags_v2():
  """测试空间连接打标签的方法"""

  # lnglat -- wkb
  df1 = pd.DataFrame(point_df[['name', 'lng', 'lat']])
  df2 = pd.DataFrame(polygon_df[[
    '板块', 'geometry_wkb']].rename(columns={'geometry_wkb': 'geometry'}))
  df3 = pd.DataFrame(res_df[[
    'name', 'lng', 'lat', 'geometry_wkb', '板块'
  ]].rename(columns={'geometry_wkb': 'geometry'}))
  assert_frame_equal(mark_tags_v2(df1, df2, '板块'), df3)

  # wkb -- wkb, with lnglat
  assert_frame_equal(
      mark_tags_v2(
          point_df[['name', 'lng', 'lat', 'geometry_wkb']].rename(
              columns={'geometry_wkb': 'geometry'}),
          polygon_df[['板块', 'geometry_wkb']].rename(
              columns={'geometry_wkb': 'geometry'}),
          col_list=['板块'],
      ),
      res_df[['name', 'lng', 'lat', 'geometry_wkb', '板块']].rename(
          columns={'geometry_wkb': 'geometry'})
  )

  # wkb -- wkb, no lnglat
  assert_frame_equal(
      mark_tags_v2(
          point_df[['name', 'geometry_wkb']].rename(
              columns={'geometry_wkb': 'geometry'}),
          polygon_df[['板块', 'geometry_wkb']].rename(
              columns={'geometry_wkb': 'geometry'}),
          col_list=['板块'],
      ),
      res_df[['name', 'geometry_wkb', '板块']].rename(
          columns={'geometry_wkb': 'geometry'})
  )

  # drop_geometry
  assert_frame_equal(
      mark_tags_v2(
          point_df[['name', 'lng', 'lat']],
          polygon_df[['板块', 'geometry_wkb']].rename(
              columns={'geometry_wkb': 'geometry'}),
          col_list=['板块'],
          drop_geometry=True
      ),
      res_df[['name', 'lng', 'lat', '板块']]
  )
  # shapely -- shapely, no lnglat
  assert_frame_equal(
      mark_tags_v2(
          point_df[['name', 'geometry_shapely']].rename(
              columns={'geometry_shapely': 'geometry'}),
          polygon_df[['板块', 'geometry_shapely']].rename(
              columns={'geometry_shapely': 'geometry'}),
          col_list=['板块'],
      ),
      res_df[['name', 'geometry_wkb', '板块']].rename(
          columns={'geometry_wkb': 'geometry'})
  )

  # shapely -- shapely, no lnglat, return shapely
  assert_frame_equal(
      mark_tags_v2(
          point_df[['name', 'geometry_shapely']].rename(
              columns={'geometry_shapely': 'geometry'}),
          polygon_df[['板块', 'geometry_shapely']].rename(
              columns={'geometry_shapely': 'geometry'}),
          col_list=['板块'],
          geometry_format='shapely'
      ),
      res_df[['name', 'geometry_shapely', '板块']].rename(
          columns={'geometry_shapely': 'geometry'})
  )


def test_mark_tags_v22():
  """测试空间连接打标签的方法"""
  mapping = {
    'lng': 'lng2',
    'lat': 'lat2',
    'geometry_wkb': 'geometry_wkb2',
    'geometry_shapely': 'geometry_shapely2',
  }
  point = point_df.rename(columns=mapping)
  polygon = polygon_df.rename(columns=mapping)
  res = res_df.rename(columns=mapping)
  # lnglat -- wkb
  assert_frame_equal(
      mark_tags_v2(
          point[['name', 'lng2', 'lat2']],
          polygon[['板块', 'geometry_wkb2']].rename(
              columns={'geometry_wkb2': 'geometry2'}),
          col_list=['板块'],
          point_lng='lng2',
          point_lat='lat2',
          point_geometry='geometry2',
          polygon_geometry='geometry2',
      ),
      res[['name', 'lng2', 'lat2', 'geometry_wkb2', '板块']].rename(
          columns={'geometry_wkb2': 'geometry2'})
  )

  # wkb -- wkb, with lnglat
  assert_frame_equal(
      mark_tags_v2(
          point[['name', 'lng2', 'lat2', 'geometry_wkb2']].rename(
              columns={'geometry_wkb2': 'geometry2'}),
          polygon[['板块', 'geometry_wkb2']].rename(
              columns={'geometry_wkb2': 'geometry2'}),
          col_list=['板块'],
          point_lng='lng2',
          point_lat='lat2',
          point_geometry='geometry2',
          polygon_geometry='geometry2',
      ),
      res[['name', 'lng2', 'lat2', 'geometry_wkb2', '板块']].rename(
          columns={'geometry_wkb2': 'geometry2'})
  )

  # wkb -- wkb, no lnglat
  assert_frame_equal(
      mark_tags_v2(
          point[['name', 'geometry_wkb2']].rename(
              columns={'geometry_wkb2': 'geometry2'}),
          polygon[['板块', 'geometry_wkb2']].rename(
              columns={'geometry_wkb2': 'geometry2'}),
          col_list=['板块'],
          point_lng='lng2',
          point_lat='lat2',
          point_geometry='geometry2',
          polygon_geometry='geometry2',
      ),
      res[['name', 'geometry_wkb2', '板块']].rename(
          columns={'geometry_wkb2': 'geometry2'})
  )

  # drop_geometry
  assert_frame_equal(
      mark_tags_v2(
          point[['name', 'lng2', 'lat2']],
          polygon[['板块', 'geometry_wkb2']].rename(
              columns={'geometry_wkb2': 'geometry2'}),
          col_list=['板块'],
          drop_geometry=True,
          point_lng='lng2',
          point_lat='lat2',
          point_geometry='geometry2',
          polygon_geometry='geometry2',
      ),
      res[['name', 'lng2', 'lat2', '板块']]
  )
  # shapely -- shapely, no lnglat
  assert_frame_equal(
      mark_tags_v2(
          point[['name', 'geometry_shapely2']].rename(
              columns={'geometry_shapely2': 'geometry2'}),
          polygon[['板块', 'geometry_shapely2']].rename(
              columns={'geometry_shapely2': 'geometry2'}),
          col_list=['板块'],
          point_lng='lng2',
          point_lat='lat2',
          point_geometry='geometry2',
          polygon_geometry='geometry2',
      ),
      res[['name', 'geometry_wkb2', '板块']].rename(
          columns={'geometry_wkb2': 'geometry2'})
  )

  # shapely -- shapely, no lnglat, return shapely
  assert_frame_equal(
      mark_tags_v2(
          point[['name', 'geometry_shapely2']].rename(
              columns={'geometry_shapely2': 'geometry2'}),
          polygon[['板块', 'geometry_shapely2']].rename(
              columns={'geometry_shapely2': 'geometry2'}),
          col_list=['板块'],
          geometry_format='shapely',
          point_lng='lng2',
          point_lat='lat2',
          point_geometry='geometry2',
          polygon_geometry='geometry2',
      ),
      res[['name', 'geometry_shapely2', '板块']].rename(
          columns={'geometry_shapely2': 'geometry2'})
  )


def test_is_xxx():
  assert is_wkt(None) is False
  assert is_wkb(None) is False
  assert is_shapely(None) is False
  assert is_geojson(None) is False

  assert is_wkt(None, na=True) is True
  assert is_wkb(None, na=True) is True
  assert is_shapely(None, na=True) is True
  assert is_geojson(None, na=True) is True

  assert is_wkt('aaa') is False
  assert is_wkb('aaa') is False
  assert is_shapely('aaa') is False
  assert is_geojson('aaa') is False

  assert is_wkt('Point(1 1)') is True
  assert is_wkb('0101000020E6100000000000000000F03F000000000000F03F') is True
  assert is_shapely(Point(1, 1)) is True
  assert is_geojson('{"type": "Point", "coordinates": [1.0, 1.0]}') is True


def test_infer_geom_format():
  assert infer_geom_format(None) == GeomFormat.unknown
  assert infer_geom_format(
      '0101000020E61000009A9999999999F13F9A9999999999F13F') == GeomFormat.wkb
  assert infer_geom_format(Point(1.1, 1.1)) == GeomFormat.shapely
  assert infer_geom_format('POINT (1.1 1.1)') == GeomFormat.wkt
  assert infer_geom_format(['POINT (1.1 1.1)']) == GeomFormat.wkt
  assert infer_geom_format(('POINT (1.1 1.1)',)) == GeomFormat.wkt
  assert infer_geom_format(pd.Series(['POINT (1.1 1.1)'])) == GeomFormat.wkt



import datetime

from ricco.util.id_number import IDNumber
from ricco.util.id_number import get_check_code

id_t = '110000200001010000'
id_f = '110000200001010001'


def test_get_check_code():
  assert get_check_code(id_t) == '0'
  assert get_check_code(id_f) == '0'


def test_id_number():
  assert IDNumber(id_t).check_code == '0'
  assert IDNumber(id_t).gender == '女'
  assert IDNumber(id_t).sex == '女'
  assert IDNumber(id_t).birthdate == datetime.datetime(2000, 1, 1)
  assert IDNumber(id_t).birth_int == 20000101
  assert IDNumber(id_t).birthday() == '2000-01-01'
  assert IDNumber.is_valid(id_t) is True
  assert IDNumber.is_valid(id_t, check_code=True) is True
  assert IDNumber.is_valid(id_f) is True
  assert IDNumber.is_valid(id_f, check_code=True) is False



from ..resource.bd_region import get_bd_region
from .base import not_empty
from .base import warn_
from .decorator import singleton


class District:
  def __init__(self):
    self.df = get_bd_region()
    self.city_list = self.city_list()
    self.region_list = self.region_list()

  @singleton
  def city_list(self):
    city_list = [
      *self.df['城市名称'].unique().tolist(),
      *self.df['城市简称'].unique().tolist(),
    ]
    return [c for c in city_list if not_empty(c)]

  @singleton
  def region_list(self):
    region_list = [
      *self.df['区县名称'].unique().tolist(),
      *self.df['区县简称'].unique().tolist(),
    ]
    return [r for r in region_list if not_empty(r)]

  def is_city(self, name) -> bool:
    """判断是否是地级市"""
    return name in self.city_list

  def is_region(self, name) -> bool:
    """判断是否是县市区"""
    return name in self.region_list

  def city_names(self, name):
    res = []
    for c in ['城市名称', '城市简称', '区县名称', '区县简称']:
      if name in self.df[c].unique():
        res = self.df[self.df[c] == name]['城市名称'].unique().tolist()
        break
    return res

  def province_names(self, name):
    res = []
    for c in ['省份名称', '省份简称', '城市名称', '城市简称', '区县名称',
              '区县简称']:
      if name in self.df[c].unique():
        res = self.df[self.df[c] == name]['省份名称'].unique().tolist()
        break
    return res

  def _get_district(self, name, type_, if_not_unique=None, warning=True):
    """
    通过地名（城市、区县）获取所在的城市名称
    Args:
      name: 城市名称
      type_: 城市、区县，可选参数city、province
      if_not_unique: 当查询到多个结果如何处理，
        - None， 默认， 返回空值
        - 'first'， 返回第一个
        - 'last'， 返回最后一个
        - 'all'， 返回所有城市的列表
      warning: 是否打印警告
    """
    assert if_not_unique in [None, 'first', 'last', 'all']
    if type_ == 'city':
      ls = self.city_names(name)
    elif type_ == 'province':
      ls = self.province_names(name)
    else:
      raise ValueError('type_ must be "city" or "province"')
    if ls:
      if len(ls) == 1:
        return ls[0]
      warn_(f'"{name}"匹配到多个城市{ls}', warning)
      if not if_not_unique:
        return
      if if_not_unique == 'first':
        return ls[0]
      if if_not_unique == 'last':
        return ls[-1]
      if if_not_unique == 'all':
        return ls

  def city(self, name, if_not_unique=None, warning=True):
    """
    通过地名（城市、区县）获取所在的城市名称
    Args:
      name: 城市名称
      if_not_unique: 当查询到多个结果如何处理，
        - None， 默认， 返回空值
        - 'first'， 返回第一个
        - 'last'， 返回最后一个
        - 'all'， 返回所有城市的列表
      warning: 是否打印警告
    """
    return self._get_district(
        name=name,
        type_='city',
        if_not_unique=if_not_unique,
        warning=warning
    )

  def province(self, name, if_not_unique=None, warning=True):
    """
    通过地名（省份、城市、区县）获取所在的省份名称
    Args:
      name: 地名
      if_not_unique: 当查询到多个结果如何处理，
        - None， 默认， 返回空值
        - 'first'， 返回第一个
        - 'last'， 返回最后一个
        - 'all'， 返回所有省份的列表
      warning: 是否打印警告
    """
    return self._get_district(
        name=name,
        type_='province',
        if_not_unique=if_not_unique,
        warning=warning
    )



import pandas as pd

from ..resource import P_BD_REGION
from ..util.base import not_empty
from ..util.decorator import singleton
from ..util.util import rstrip_d0

EX = ['新城区']


@singleton
def get_bd_region() -> pd.DataFrame:
  """获取bd_region表并转为宽表；同时保存为全局变量，避免多次IO"""
  df = pd.read_csv(P_BD_REGION)
  for c in ['parent_id', 'id']:
    df[c] = df[c].astype(str)
    df[c] = df[c].apply(rstrip_d0)
  df = df[df['parent_id'].notna()]
  df = df[df['level_type'].isin([1, 2, 3])]
  df['level_type'] = df['level_type'].replace(
      to_replace={1: '省份', 2: '城市', 3: '区县'}
  )
  # 区县
  df_region = df[df['level_type'] == '区县']
  df_region = df_region[['id', 'name', 'short_name', 'parent_id']]
  df_region.columns = ['区县id', '区县名称', '区县简称', '城市id']
  # 城市
  df_city = df[df['level_type'] == '城市']
  df_city = df_city[['id', 'name', 'short_name', 'parent_id']]
  df_city.columns = ['城市id', '城市名称', '城市简称', '省份id']
  # 省份
  df_prov = df[df['level_type'] == '省份']
  df_prov = df_prov[['id', 'name', 'short_name']]
  df_prov.columns = ['省份id', '省份名称', '省份简称']
  # 合并
  df_res = df_region.merge(df_city, how='outer', on='城市id')
  return df_res.merge(df_prov, how='outer', on='省份id')


@singleton
def cities():
  """城市列表，按照长度排序"""
  ls = get_bd_region()['城市名称'].unique().tolist()
  ls = [i for i in ls if not_empty(i)]
  return sorted(ls, key=lambda x: len(x), reverse=True)


@singleton
def regions():
  """区县列表，按照长度排序"""
  ls = get_bd_region()['区县名称'].unique().tolist()
  ls = [i for i in ls if not_empty(i) and i not in EX]
  return sorted(ls, key=lambda x: len(x), reverse=True)


@singleton
def city_region_list():
  """城市和区县列表，按照长度排序"""
  df = get_bd_region()
  ls = [*df['城市名称'].unique().tolist(), *df['区县名称'].unique().tolist()]
  ls = [i for i in ls if not_empty(i) and i not in EX]
  return sorted(ls, key=lambda x: len(x), reverse=True)



import logging
import uuid
import warnings

import pandas as pd
from tqdm import tqdm

from ..etl.transformer import create_columns
from ..util.decorator import check_null
from .amap import get_address_amap
from .amap import get_place_amap
from .baidu import get_address_baidu
from .baidu import get_place_baidu
from .util import DEFAULT_RES


def geocode(*,
            address,
            city,
            source,
            srs='wgs84',
            key_baidu=None,
            key_amap=None):
  """
  Args:
    address: 地址或项目名称关键词
    city: 城市
    source: 接口选择。baidu/amap：百度或高德的地理编码接口，baidu_poi/amap_poi：百度或高德的地点检索接口
    srs: 返回的坐标系，可选wgs84, bd09, gcj02，默认wgs84
    key_baidu: 百度接口的key，公共key失效后可自行传入
    key_amap: 高德接口的key，公共key失效后可自行传入
  """
  assert source in ('baidu', 'baidu_poi', 'amap', 'amap_poi')
  if source == 'baidu':
    return get_address_baidu(address=address, city=city, srs=srs, key=key_baidu)
  if source == 'baidu_poi':
    return get_place_baidu(address=address, city=city, srs=srs, key=key_baidu)
  if source == 'amap':
    return get_address_amap(address=address, city=city, srs=srs, key=key_amap)
  if source == 'amap_poi':
    return get_place_amap(address=address, city=city, srs=srs, key=key_amap)


@check_null(default_rv=DEFAULT_RES)
def geocode_best_poi(
    address,
    city,
    sig=80,
    address_geocode=False,
    srs='wgs84',
    key_baidu=None,
    key_amap=None):
  """
  获取最优的地点检索结果
  Args:
    address: POI名称
    city: 城市
    sig: 解析结果相似度，默认为80，范围0-100
    address_geocode: 是否使用地理编码补全，默认False
    srs: 返回的坐标系，可选wgs84, bd09, gcj02，默认wgs84
    key_baidu: 百度接口的key，公共key失效后可自行传入
    key_amap: 高德接口的key，公共key失效后可自行传入
  """
  try:
    ar = get_place_amap(address, city, srs=srs, key=key_amap)
    if ar['score'] >= 90:
      return ar
    br = get_place_baidu(address, city, srs=srs, key=key_baidu)
    if br['score'] >= sig or ar['score'] >= sig:
      return br if br['score'] > ar['score'] else ar
    if address_geocode:
      br = get_address_baidu(address, city, srs=srs, key=key_baidu)
      if br['score'] >= 90:
        return br
      return get_address_amap(address, city, srs=srs, key=key_amap)
  except Exception as e:
    warnings.warn(f'结果获取失败：【{city}】：【{address}】,{e}')
    return DEFAULT_RES


@check_null(default_rv=DEFAULT_RES)
def geocode_best_address(address, city, srs='wgs84', key_baidu=None,
                         key_amap=None):
  """获取最优的地理编码结果"""
  try:
    br = get_address_baidu(address, city, srs=srs, key=key_baidu)
    if br['score'] >= 90:
      return br
    return get_address_amap(address, city, srs=srs, key=key_amap)
  except Exception as e:
    warnings.warn(f'结果获取失败：【{city}】：【{address}】,{e}')
    return DEFAULT_RES


def geocode_best(
    address,
    city,
    address_type: str,
    sig: int = 80,
    srs: str = 'wgs84',
    address_geocode: bool = False,
    **kwargs):
  assert address_type in ('poi', 'address'), 'address_type可选参数为poi或address'
  if address_type == 'poi':
    return geocode_best_poi(
        address, city, sig=sig, address_geocode=address_geocode, srs=srs,
        **kwargs
    )
  return geocode_best_address(address, city, srs=srs, **kwargs)


@check_null(default_rv=DEFAULT_RES)
def geocode_with_memory_cache(
    address, city, address_type: str,
    sig: int = 80, srs: str = 'wgs84',
    address_geocode: bool = False,
    key_baidu=None, key_amap=None,
    ignore_existing=True, lng=None, lat=None, memory_cache: bool = True):
  """geocoding（使用内存加速）"""
  if memory_cache:
    global __GLOBAL_CACHE
    if '__GLOBAL_CACHE' not in globals():
      __GLOBAL_CACHE = {}
    if city in __GLOBAL_CACHE:
      if address in __GLOBAL_CACHE[city]:
        return __GLOBAL_CACHE[city][address]

  if ignore_existing and all([lng, lat]):
    rv = DEFAULT_RES.copy()
    rv['lng'] = lng
    rv['lat'] = lat
    return rv
  res = geocode_best(
      address=address, city=city, sig=sig,
      address_type=address_type, address_geocode=address_geocode,
      srs=srs, key_baidu=key_baidu, key_amap=key_amap
  )
  if memory_cache:
    if city not in __GLOBAL_CACHE:
      __GLOBAL_CACHE[city] = {}
    __GLOBAL_CACHE[city][address] = res
  return res


def geocode_df(df: pd.DataFrame,
               by: str,
               city: (str, list),
               address_type: str,
               sig: int = 80,
               address_geocode: bool = False,
               with_detail: bool = True,
               srs: str = 'wgs84',
               **kwargs):
  """
  对dataframe进行geocoding
  Args:
    df:
    by: 项目名称所在的列
    city: 城市，list表示城市列如['城市']， str表示全部使用该城市如'上海'
    address_type: 'poi'或'address'，不同的地理类型对应不同的接口，精确度不同
    sig: 解析结果相似度，默认为80，范围为0-100，仅当address_type为poi时生效
    address_geocode: poi解析时是否使用地理编码补全，默认False，仅当address_type为poi时生效
    with_detail: 是否返回详情信息，包括返回值、得分、接口来源等
    srs: 返回的坐标系，可选wgs84, bd09, gcj02，默认wgs84
    key_baidu: 百度接口的key，公共key失效后可自行传入
    key_amap: 高德接口的key，公共key失效后可自行传入
  """
  assert address_type in ('poi', 'address'), 'address_type可选参数为poi或address'
  assert isinstance(city, (str, list))
  assert 'lng' not in df, '原数据中不能有lng列'
  assert 'lat' not in df, '原数据中不能有lat列'

  __c_city = uuid.uuid1()
  on = [__c_city, by]

  df[__c_city] = city if isinstance(city, str) else df[city]

  _df = df[on].drop_duplicates().reset_index(drop=True)
  base_cols = ['lng', 'lat']
  if with_detail:
    base_cols.extend(['rv', 'score', 'source'])
  _df = create_columns(_df, base_cols)
  for i in tqdm(_df.index):
    _address = None
    try:
      _address, _city = _df[by][i], _df[__c_city][i]
      rv = geocode_best(
          address=_address, city=_city, sig=sig,
          address_type=address_type, address_geocode=address_geocode,
          srs=srs,
          **kwargs
      )
      for c in base_cols:
        _df.loc[(_df[__c_city] == _city) & (_df[by] == _address), c] = rv[c]
    except Exception as e:
      logging.warning(f'{e},{_address}')
  df = df.merge(_df, on=on, how='left')
  del df[__c_city]
  return df


def geocode_v2(df: pd.DataFrame,
               by: str,
               city: (str, list),
               address_type: str,
               sig: int = 80,
               address_geocode: bool = False,
               with_detail: bool = True,
               srs: str = 'wgs84',
               ignore_existing: bool = True,
               memory_cache: bool = True,
               **kwargs):
  """
  对dataframe进行geocoding
  Args:
    df:
    by: 项目名称所在的列
    city: 城市，list表示城市列如['城市']， str表示全部使用该城市如'上海'
    address_type: 'poi'或'address'，不同的地理类型对应不同的接口，精确度不同
    sig: 解析结果相似度，默认为80，范围为0-100，仅当address_type为poi时生效
    address_geocode: poi解析时是否使用地理编码补全，默认False，仅当address_type为poi时生效
    with_detail: 是否返回详情信息，包括返回值、得分、接口来源等
    srs: 返回的坐标系，可选wgs84, bd09, gcj02，默认wgs84
    ignore_existing: 是否忽略已有的经纬度，默认True
    key_baidu: 百度接口的key，公共key失效后可自行传入
    key_amap: 高德接口的key，公共key失效后可自行传入
    memory_cache: 是否使用内存缓存，默认True
  """
  assert address_type in ('poi', 'address'), 'address_type可选参数为poi或address'
  assert isinstance(city, (str, list))
  assert df.index.is_unique, 'DataFrame索引列必须唯一'

  __c_city = uuid.uuid1()
  df[__c_city] = city if isinstance(city, str) else df[city]

  base_cols = ['lng', 'lat']
  if with_detail:
    base_cols.extend(['rv', 'score', 'source'])
  df = create_columns(df, base_cols)

  for i in tqdm(df.index):
    _address, _city = df[by][i], df[__c_city][i]
    _lng, _lat = df['lng'][i], df['lat'][i]
    try:
      rv = geocode_with_memory_cache(
          address=_address, city=_city, sig=sig,
          address_type=address_type, address_geocode=address_geocode,
          srs=srs,
          ignore_existing=ignore_existing, lng=_lng, lat=_lat,
          memory_cache=memory_cache, **kwargs
      )
      for c in base_cols:
        df.loc[(df[__c_city] == _city) & (df[by] == _address), c] = rv[c]
    except Exception as e:
      logging.warning(f'{e},{_address}')
  del df[__c_city]
  return df



import datetime
import logging
import os
import zipfile

from .base import ensure_list


def ext(filepath):
  """扩展名"""
  return os.path.splitext(filepath)[1]


def extension(filepath):
  """扩展名"""
  return os.path.splitext(filepath)[1]


def fn(filepath):
  """路径及文件名（不含扩展名）"""
  return os.path.splitext(filepath)[0]


def path_name(filepath):
  """路径及文件名（不含扩展名）"""
  return os.path.splitext(filepath)[0]


def split_path(filepath, abspath=False):
  """将文件路径拆分为文件夹路径、文件名、扩展名三部分"""
  if abspath:
    filepath = os.path.abspath(filepath)
  basename = os.path.basename(filepath)
  dir_path = os.path.dirname(filepath)
  return dir_path, path_name(basename), extension(basename)


def ensure_dir(dirpath: str):
  if not dirpath.endswith('/'):
    return dirpath + '/'
  return dirpath


def remove_dir(filepath):
  """除某一目录下的所有文件或文件夹"""
  import shutil
  del_list = os.listdir(filepath)
  for f in del_list:
    file_path = os.path.join(filepath, f)
    if os.path.isfile(file_path):
      os.remove(file_path)
    elif os.path.isdir(file_path):
      shutil.rmtree(file_path)
  shutil.rmtree(filepath)


def find_undefined_dirs(dirpath):
  """获取不存在的目录列表"""
  dirs = []
  dirpath = os.path.abspath(dirpath)
  while not os.path.exists(dirpath):
    dirs.append(dirpath)
    dirpath = os.path.dirname(dirpath)
  dirs.reverse()
  return dirs


def ensure_dirpath_exist(filepath):
  """确保目录存在，不存在则创建"""
  dir_path = os.path.dirname(filepath)
  if os.path.exists(dir_path):
    return
  undefined_dirs = find_undefined_dirs(dir_path)
  for p in undefined_dirs:
    logging.warning(f'目录："{p}"不存在，已创建')
    os.makedirs(p)


def dir2zip(filepath, overwrite=False, delete_origin=False):
  """压缩文件夹"""
  zfn = f'{filepath}.zip'
  if os.path.exists(zfn):
    if overwrite:
      os.remove(zfn)
    else:
      raise FileExistsError(f'目标文件{zfn}已存在')
  z = zipfile.ZipFile(zfn, 'w', zipfile.ZIP_DEFLATED)
  for dirpath, _, filenames in os.walk(filepath):
    for filename in filenames:
      filepath_out = os.path.join(dirpath, filename)
      filepath_in = os.path.join(os.path.split(dirpath)[-1], filename)
      z.write(filepath_out, arcname=filepath_in)
  z.close()
  if delete_origin:
    print(f'delete {filepath}')
    remove_dir(filepath)


def get_file_counts(dir_path):
  """获取文件夹中文件的数量"""
  num = 0
  for dirpath, dirnames, filenames in os.walk(dir_path):
    for filename in filenames:
      if filename != '.DS_Store':
        num += 1
  return num


def rm_scratch_file(dir_path, days, recursive=False, rm_hidden_file=False):
  """删除指定天数前修改过的文件"""
  dirs = [
    '/bin', '/sbin', '/usr', '/etc', '/dev', '/Applications', '/Library',
    '/Network', '/System', '/Volumes', '/cores', '/private'
  ]
  for pre in dirs:
    if os.path.abspath(dir_path).startswith(fr'{pre}'):
      raise ValueError('系统目录下的文件不可删除')
  if dir_path == '/':
    raise ValueError('不可指定根目录')
  timeline = datetime.datetime.now() - datetime.timedelta(days)
  # 删除过期文件
  for file in dir_iter(dir_path, abspath=True, recursive=recursive,
                       ignore_hidden_files=not rm_hidden_file):
    m_time = os.path.getmtime(file)
    m_time = datetime.datetime.fromtimestamp(m_time)
    if m_time <= timeline:
      os.remove(file)
      logging.warning(f'已删除文件：{file}')
  # 删除空文件夹
  for dirpath, dirnames, _ in os.walk(dir_path):
    for dirname in dirnames:
      dir_full_path = os.path.join(dirpath, dirname)
      num = get_file_counts(dir_full_path)
      if num == 0:
        remove_dir(dir_full_path)
        logging.warning(f'已删除空文件夹：{dir_full_path}')


def dir_iter(root,
             exts: (list, str) = None,
             abspath=False,
             recursive=False,
             ignore_hidden_files=True):
  """
  文件夹中的文件路径生成器，用于遍历文件夹中的文件
  Args:
    root: 文件目录
    exts: 文件扩展名，不指定则返回所有文件
    abspath: 是否返回绝对路径
    recursive: 是否循环遍历更深层级的文件，默认只返回当前目录下的文件
    ignore_hidden_files: 是否忽略隐藏文件
  """
  if exts:
    exts = ensure_list(exts)

  for dirpath, _, filenames in os.walk(root):
    for _name in filenames:
      if ignore_hidden_files and _name.startswith('.'):
        continue
      filepath = os.path.join(dirpath, _name)
      filepath = os.path.abspath(filepath) if abspath else filepath
      # 不符合扩展名要求忽略
      if exts and extension(_name) not in exts:
        continue
      # 仅需要当前目录时，dirpath和root不相同的过滤掉
      if not recursive and dirpath != root:
        continue
      yield filepath


def dir_iter_list(root,
                  exts: (list, str) = None,
                  abspath=False,
                  recursive=False,
                  reverse=False):
  """
  文件夹中的文件路径列表
  Args:
    root: 文件目录
    exts: 文件扩展名，不指定则返回所有文件
    abspath: 是否返回绝对路径
    recursive: 是否循环遍历更深层级的文件，默认只返回当前目录下的文件
    reverse: 路径列表是否倒序
  """
  path_list = []
  for p in dir_iter(root, exts, abspath, recursive, ignore_hidden_files=True):
    path_list.append(p)
  return sorted(path_list, reverse=reverse)


def single_ext(path_list):
  ext_list = list(set([extension(p) for p in path_list]))
  if len(ext_list) == 1:
    return ext_list[0]



from ricco.util.district import District


def test_district():
  ds = District()
  assert ds.city(None) == None
  assert ds.city('') == None
  assert ds.city('单县') == '菏泽市'
  assert ds.city('杨浦区') == '上海市'
  assert ds.province(None) == None
  assert ds.province('') == None
  assert ds.province('滕州市') == '山东省'



import pandas as pd

from ricco.etl.entropy import EntropyClass
from ricco.etl.entropy import entropy
from ricco.etl.entropy import rescale
from ricco.etl.entropy import standard_e
from ricco.etl.entropy import standard_e_neg


def test_standard_e():
  df = pd.DataFrame({
      'name': ['a', 'b', 'c'],
      'value_1': [1, 2, 3],
      'value_2': [-1, -2, -3],
  })
  assert min(list(standard_e(df, ['value_1'])['value_1'])) == 0
  assert max(list(standard_e(df, ['value_1'])['value_1'])) == 1
  assert list(standard_e(df, ['value_1'])['value_1']) == [0.0, 0.5, 1.0]
  assert list(standard_e(df, ['value_1'])['value_2']) == [-1, -2, -3]


def test_standard_e_neg():
  df = pd.DataFrame({
      'name': ['a', 'b', 'c'],
      'value_1': [1, 2, 3],
      'value_2': [-1, -2, -3],
  })
  assert min(list(standard_e_neg(df, ['value_2'])['value_2'])) == 0
  assert max(list(standard_e_neg(df, ['value_2'])['value_2'])) == 1
  assert list(standard_e_neg(df, ['value_2'])['value_2']) == [0.0, 0.5, 1.0]
  assert list(standard_e_neg(df, ['value_2'])['value_1']) == [1, 2, 3]


def test_entropy():
  df = pd.DataFrame({
      'name': ['a', 'b', 'c'],
      'value_1': [1, 2, 3],
      'value_2': [-1, -2, -3],
  })
  df_s = standard_e(df, ['value_1'])
  df_s = standard_e_neg(df_s, ['value_2'])
  x, weight = entropy(df_s, columns=['value_1', 'value_2'])
  assert weight['value_1'] == weight['value_2']
  assert list(x) == [0, 1 / 3, 2 / 3]


def test_rescale():
  df = pd.DataFrame({
      'name': ['a', 'b', 'c'],
      'value_1': [1, 2, 3],
      'value_2': [-1, -2, -3],
  })
  assert list(rescale(df, 'value_1')) == [0, 50, 100]
  assert max(rescale(df, 'value_1', score_range=(50, 70))) == 70
  assert min(rescale(df, 'value_1', score_range=(50, 70))) == 50
  assert min(rescale(df, 'value_2', score_range=(20, 30))) == 20
  assert max(rescale(df, 'value_2', score_range=(20, 30))) == 30
  assert min(rescale(df, 'value_1', score_range=(-20, 50))) == -20
  assert max(rescale(df, 'value_1', score_range=(-20, 50))) == 50


def test_EntropyClass():
  df = pd.DataFrame({
      'name': ['a', 'b', 'c'],
      'value_1': [1, 2, 3],
      'value_2': [-1, -2, -3],
  })
  Entropy = EntropyClass(df, ['value_1', 'value_2'], ['value_2'], '得分')
  assert list(Entropy.entropy_res) == [0, 1 / 3, 2 / 3]
  assert Entropy.entropy_weight['value_1'] == Entropy.entropy_weight['value_2']
  assert "得分" in Entropy.entropy_df.columns
  assert list(Entropy.entropy_df['得分']) == [0, 1 / 3, 2 / 3]



import datetime as dt
import re
from datetime import datetime
from datetime import timedelta

import pandas as pd
from dateutil.relativedelta import relativedelta

from .decorator import check_null
from .decorator import to_str


@check_null()
def auto2date(string, errors='raise'):
  """自动检查格式并输出日期"""
  if string in ['', None]:
    return
  string = str(string)

  if '时' in string:
    _hour = re.findall('\d+时', string)[0]
    string = string.split(_hour)[0]
  if all([i in string for i in ['年', '月', '日']]):
    return pd.to_datetime(string, format='%Y年%m月%d日')
  if all([i in string for i in ['年', '月']]):
    return pd.to_datetime(string, format='%Y年%m月')
  if '年' in string:
    return pd.to_datetime(string, format='%Y年')
  return pd.to_datetime(string, errors=errors)


def is_valid_date(string, na=False):
  """判断是否是一个有效的日期字符串"""
  if string in ['', None]:
    return na
  try:
    auto2date(string)
    return True
  except Exception:
    return False


@check_null()
def excel2date(dates, date_type='str'):
  """
  excel的数字样式时间格式转日期格式
  Args:
    dates: 传入excel的日期格式，或ymd的日期格式
    date_type: 返回日期类型，str or date
  """
  from xlrd import xldate_as_datetime
  dates = str(dates)
  if re.match('^\d{4,5}$', dates):
    _date = datetime.strftime(xldate_as_datetime(int(dates), 0), '%Y-%m-%d')
  elif is_valid_date(dates):
    _date = datetime.strftime(auto2date(dates), '%Y-%m-%d')
  else:
    return

  if date_type in ('string', 'str', str):
    return _date
  elif date_type in ('date', 'datetime'):
    return datetime.strptime(_date, '%Y-%m-%d')
  else:
    raise ValueError('date_type参数错误，可选参数为str或date')


class DT:

  def __init__(self,
               date: (str, dt.date, dt.datetime) = None,
               format='%Y-%m-%d',
               dst_format=None):
    """
    DT日期类
    Args:
      date: 初始化支持字符串、datetime.date和datetime.datetime格式作为基准日期。不指定时，默认基准日期今天。
      format: 当初始化为字符串时，需要输入format指定日期字符串格式。
      dst_format: 可选。未指定时，类方法输出结果为datetime.date日期格式。指定dst_format时，输出结果为符合dst_format的字符串格式。
    """
    if date:
      if isinstance(date, str):
        self.date = dt.datetime.strptime(date, format).date()
      elif isinstance(date, dt.datetime):
        self.date = date.date()
      elif isinstance(date, dt.date):
        self.date = date
      else:
        raise TypeError(
            '无法识别的类型，date只能为字符串、datetime.date或datetime.datetime类型')
    else:
      self.date = dt.date.today()
    self.dst_format = dst_format

  @property
  @to_str
  def today(self):
    return self.date

  @property
  @to_str
  def tomorrow(self):
    return self.date_move(days=1)

  @property
  @to_str
  def yesterday(self):
    return self.date_move(days=-1)

  @property
  @to_str
  def the_day_after_tomorrow(self):
    return self.date_move(days=2)

  @property
  @to_str
  def the_day_before_yesterday(self):
    return self.date_move(days=-2)

  @property
  @to_str
  def one_year_ago(self):
    return self.date_move(years=-1)

  @property
  @to_str
  def first_day_of_this_month(self):
    return self.date.replace(day=1)

  @property
  @to_str
  def last_day_of_this_month(self):
    return self.date_move(months=1).replace(day=1) + relativedelta(days=-1)

  @property
  @to_str
  def day_half_year_ago(self):
    return self.date_move(months=-6)

  def date_move(self, years=0, months=0, days=0):
    return self.date + relativedelta(years=years, months=months, days=days)

  @property
  @to_str
  def monday_of_this_week(self):
    weekday = self.date.isoweekday()
    if weekday == 1:
      return self.date
    return self.date - timedelta(weekday - 1)



import pandas as pd

from ..util.util import is_digit
from .transformer import round_by_columns

class_dic = {
  'count': '计数',
  'mean': '均值',
  'std': '标准差',
  'min': '最小值',
  'max': '最大值',
  '50%': '中位数',
}

skew_skurt = '''偏度（Skewness）

用来描述数据分布的对称性，正态分布的偏度为0。
计算数据样本的偏度，当偏度<0时，称为负偏，数据出现左侧长尾；
当偏度>0时，称为正偏，数据出现右侧长尾；当偏度为0时，表示数据相对均匀的分布在平均值两侧，
不一定是绝对的对称分布，此时要与正态分布偏度为0的情况进行区分。

当偏度绝对值过大时，长尾的一侧出现极端值的可能性较高。


峰度(Kurtosis)

用来描述数据分布陡峭或是平滑的情况。正态分布的峰度为3，
峰度越大，代表分布越陡峭，尾部越厚；
峰度越小，分布越平滑。
很多情况下，为方便计算，将峰度值－3，因此正态分布的峰度变为0，方便比较。

在方差相同的情况下，峰度越大，存在极端值的可能性越高。
'''


def describe_series(df: pd.DataFrame, col):
  """数值型列的描述性统计"""
  length = df.shape[0]
  df_desc = pd.DataFrame(df[col].describe().reset_index())
  df_desc = df_desc.rename(columns={'index': '分类', col: '值'})
  df_desc['分类'] = df_desc['分类'].replace(to_replace=class_dic)
  # 偏度系数和峰度系数
  skew_add = pd.DataFrame({
    '分类': '偏度系数',
    '值': [df[col].skew()],
  })
  kurt_add = pd.DataFrame({
    '分类': '峰度系数',
    '值': [df[col].kurt() - 3],
  })
  # 缺失数量和缺失率
  null_num = length - df_desc.loc[df_desc['分类'] == '计数', '值'][0]
  null_rate = null_num / length
  null_add = pd.DataFrame({'分类': '缺失数', '值': [null_num]})
  null_rate_add = pd.DataFrame({'分类': '缺失率', '值': [null_rate]})
  df_desc = pd.concat(
      [df_desc, skew_add, kurt_add, null_add, null_rate_add],
      ignore_index=True,
  )
  df_desc = round_by_columns(df_desc, ['值'])
  return df_desc


def describe_object(df, col):
  desc = pd.DataFrame(df[col].value_counts().reset_index())
  if desc.shape[0] > 20:
    name = f'{col}_Top15'
    desc = desc.rename(columns={'index': name, col: '计数'})
    desc[name] = desc[name].replace(to_replace=class_dic)
    desc = desc.head(15)
  else:
    desc = desc.rename(columns={'index': col, col: '计数'})
    desc[col] = desc[col].replace(to_replace=class_dic)
  return desc


def describe_date(df: pd.DataFrame, col):
  df = df[df[col].notna()].reset_index(drop=True)
  year_num = df[col].unique().shape[0]
  if year_num >= 3:
    df[col] = df[col].dt.strftime('%Y年')
    desc = pd.DataFrame(
        df[col].value_counts().sort_index().reset_index()
    )
    desc.columns = ['年份', '计数']
  else:
    df[col] = df[col].dt.strftime('%Y-%m')
    desc = pd.DataFrame(
        df[col].value_counts().sort_index().reset_index()
    )
    desc.columns = ['月份', '计数']
  return desc


def describe_auto(df: pd.DataFrame, col: str):
  _num = df[col].unique().shape[0]
  if df[col].dtype in (float, int) and _num > 10:
    return describe_series(df, col)
  if df[col].dtype not in (float, int):
    try:
      df[col] = pd.to_datetime(df[col])
    except ValueError:
      pass
  if df[col].dtype == 'datetime64[ns]':
    return describe_date(df, col)
  return describe_object(df, col)


def suspect_series_type(df, col):
  _r = 0.95
  length = df[df[col].notna()].shape[0]
  digit_num = df[df[col].apply(is_digit)].shape[0]
  if _r <= (rates := digit_num / length) < 1:
    return f'"{col}"列有超过{int(rates * 100)}%的值为"数值型"'

  date_num = df[pd.to_datetime(df[col], errors='coerce').notna()].shape[0]
  if _r <= (rates := date_num / length) < 1:
    return f' {col} 列有超过{int(rates * 100)}%的值为"日期格式"'



import json
import re
import sys
import warnings
from typing import List

import geojson
import numpy as np
import pandas as pd
from shapely import wkb
from shapely import wkt
from shapely.errors import GeometryTypeError
from shapely.errors import WKBReadingError
from shapely.geometry import MultiLineString
from shapely.geometry import MultiPolygon
from shapely.geometry import Point
from shapely.geometry import Polygon
from shapely.geometry import shape
from shapely.geometry.base import BaseGeometry
from shapely.geos import WKTReadingError

from ..util.base import ensure_list
from ..util.base import is_empty
from ..util.base import not_empty
from ..util.decorator import check_null
from ..util.decorator import check_shapely
from ..util.decorator import check_str
from ..util.util import is_hex
from ..util.util import isinstance_in_list


class GeomFormat:
  wkb = 'wkb'
  wkt = 'wkt'
  shapely = 'shapely'
  geojson = 'geojson'
  unknown = 'unknown'


def crs_sh2000():
  """测绘院（上海2000）crs信息"""
  from ..resource.crs import CRS_SH2000
  return CRS_SH2000


def epsg_from_lnglat(lng, lat=0):
  """根据经纬度确定EPSG代码"""
  # 计算 UTM 区域代码
  lng = ensure_list(lng)
  lng = [i for i in lng if not_empty(i)]
  lng = np.median(lng)
  utm_band = str(int((np.floor((lng + 180) / 6) % 60) + 1)).zfill(2)
  return int(f'326{utm_band}' if lat >= 0 else f'327{utm_band}')


def lng_from_city(city: str):
  """获取城市所在的经度"""
  from ..resource.epsg_code import CITY_POINT
  assert len(city) >= 2, '城市名称过短'
  if city in CITY_POINT:
    return CITY_POINT[city]['lng']
  else:
    for _c in CITY_POINT.keys():
      if city in _c:
        return CITY_POINT[_c]['lng']
  warnings.warn(f'请补充"{city}"的epsg信息，默认返回经度113.0')
  return 113.0


def get_epsg(city: str):
  """根据城市查询epsg代码，用于投影"""
  return epsg_from_lnglat(lng_from_city(city))


def projection_lnglat(lnglat, crs_from, crs_to):
  """对经纬度进行投影"""
  from pyproj import Transformer
  transformer = Transformer.from_crs(crs_from, crs_to)
  return transformer.transform(xx=lnglat[1], yy=lnglat[0])


@check_str
def wkb_loads(x: str, hex=True):
  try:
    return wkb.loads(x, hex=hex)
  except (AttributeError, WKBReadingError) as e:
    warnings.warn(f'{e}, 【{x}】')


@check_shapely
def wkb_dumps(x: BaseGeometry, hex=True, srid=4326):
  try:
    return wkb.dumps(x, hex=hex, srid=srid)
  except AttributeError as e:
    warnings.warn(f'{e}, 【{x}】')


@check_str
def wkt_loads(x: str):
  try:
    return wkt.loads(x)
  except (AttributeError, WKTReadingError, TypeError) as e:
    warnings.warn(f'{e}, 【{x}】')


@check_shapely
def wkt_dumps(x: BaseGeometry):
  try:
    return wkt.dumps(x)
  except AttributeError as e:
    warnings.warn(f'{e}, 【{x}】')


@check_null()
def geojson_loads(x: (str, dict)):
  """geojson文本形式转为shapely格式"""
  from simplejson.errors import JSONDecodeError
  if not isinstance(x, (str, dict)):
    warnings.warn(f'TypeError:【{x}】')
    return
  try:
    geom = shape(geojson.loads(x)) if isinstance(x, str) else shape(x)
    if geom.is_empty:
      return
    return geom
  except (JSONDecodeError, AttributeError, GeometryTypeError, ValueError) as e:
    warnings.warn(f'{e}, 【{x}】')


@check_shapely
def geojson_dumps(x: BaseGeometry):
  """shapely转为geojson文本格式"""
  try:
    geom = geojson.Feature(geometry=x)
    return json.dumps(geom.geometry)
  except TypeError as e:
    warnings.warn(f'{e}, 【{x}】')


def is_shapely(x, na=False) -> bool:
  """判断是否为shapely格式"""
  if is_empty(x):
    return na
  if isinstance(x, BaseGeometry):
    return True
  return False


def is_wkb(x, na=False) -> bool:
  """判断是否为wkb格式"""
  if is_empty(x):
    return na
  if not isinstance(x, str) or not is_hex(x):
    return False
  try:
    wkb.loads(x, hex=True)
    return True
  except WKBReadingError:
    return False


def is_wkt(x, na=False) -> bool:
  """判断是否为wkt格式"""
  if is_empty(x):
    return na
  if not isinstance(x, str) or not re.match('^[MPL]', x.lstrip(' (')):
    return False
  try:
    wkt.loads(x)
    return True
  except WKTReadingError:
    return False


def is_geojson(x, na=False) -> bool:
  """判断是否为geojson格式"""
  from simplejson.errors import JSONDecodeError

  if is_empty(x):
    return na
  if not isinstance(x, (str, dict)):
    return False
  try:
    geom = shape(geojson.loads(x)) if isinstance(x, str) else shape(x)
    if geom.is_empty:
      return False
    return True
  except (JSONDecodeError, AttributeError, GeometryTypeError, TypeError):
    return False


def _infer_geom_format(x):
  """推断geometry格式"""
  if is_shapely(x):
    return GeomFormat.shapely
  if is_wkb(x):
    return GeomFormat.wkb
  if is_wkt(x):
    return GeomFormat.wkt
  if is_geojson(x):
    return GeomFormat.geojson
  return GeomFormat.unknown


def infer_geom_format(series: (str, list, tuple, pd.Series, BaseGeometry)):
  """推断geometry格式"""
  if is_empty(series):
    return GeomFormat.unknown
  assert isinstance(series, (str, list, tuple, pd.Series, BaseGeometry))
  if isinstance(series, (list, tuple, pd.Series)):
    for i in series:
      if not_empty(i):
        return _infer_geom_format(i)
  return _infer_geom_format(series)


def ensure_multi_geom(geom):
  """将LineString和Polygon转为multi格式"""
  if geom.geom_type == 'LineString':
    return MultiLineString([geom])
  if geom.geom_type == 'Polygon':
    return MultiPolygon([geom])
  return geom


def multiline2multipolygon(multiline_shapely: MultiLineString, force=False):
  """multiline转为multipolygon，直接首尾相连"""
  coords = []
  for line in multiline_shapely.geoms:
    lngs, lats = line.xy
    for i in range(len(lngs)):
      lng, lat = lngs[i], lats[i]
      point = Point((lng, lat))
      if coords and point == coords[-1]:
        continue
      coords.append(point)
  if len(coords) <= 2:
    return
  if not force and coords[0].distance(coords[-1]) >= 0.000001:
    return
  try:
    return MultiPolygon([Polygon(coords)])
  except Exception as e:
    warnings.warn(f'{e}，{multiline_shapely}')
    return


def get_inner_point(polygon: Polygon, within=True):
  """返回面内的一个点，默认返回中心点，当中心点不在面内则返回面内一个点"""
  if is_empty(polygon):
    return
  point = polygon.centroid
  if not polygon.is_valid:
    polygon = polygon.buffer(0.000001)
  if polygon.contains(point) or not within:
    return point
  return polygon.representative_point()


def is_valid_lnglat(lng, lat):
  """判断经纬度是否有效"""
  return -180 <= lng <= 180 and -90 <= lat <= 90


def auto_loads(x) -> BaseGeometry:
  """自动识别地理格式并转换为shapely格式"""
  geom_format = infer_geom_format(x)
  assert geom_format != GeomFormat.unknown, '未知的地理格式'
  if geom_format == 'shapely':
    return x
  return getattr(sys.modules[__name__], f'{geom_format}_loads')(x)


def dumps2x(x: BaseGeometry, geom_format):
  """转换geometry格式"""
  assert geom_format in ('wkb', 'wkt', 'shapely', 'geojson'), '未知的地理格式'
  x = auto_loads(x)
  if geom_format == 'shapely':
    return x
  return getattr(sys.modules[__name__], f'{geom_format}_dumps')(x)


def ensure_lnglat(lnglat) -> tuple:
  if is_empty(lnglat):
    return np.nan, np.nan
  if isinstance(lnglat, (list, tuple)):
    assert isinstance_in_list(lnglat, (float, int)), '数据类型错误，经度和纬度都应该为数值型'
    assert is_valid_lnglat(*lnglat), '经纬度超出范围'
    return tuple(lnglat)
  if geom := auto_loads(lnglat):
    return geom.centroid.x, geom.centroid.y
  raise ValueError('未知的地理类型')


def distance(
    p1: (tuple, str),
    p2: (tuple, str),
    city: str = None,
    epsg_from: int = 4326,
    epsg_to: (str, int) = None):
  """
  计算两个点（经度，纬度）之间的距离，单位：米
  Args:
    p1: 点位1，经纬度或geometry
    p2: 点位2，经纬度或geometry
    city: 所在城市，用于投影
    epsg_from: epsg代码
    epsg_to: 投影epsg代码
  """
  p1, p2 = ensure_lnglat(p1), ensure_lnglat(p2)
  if not epsg_to:
    epsg_to = get_epsg(city) if city else epsg_from_lnglat([p1[0], p2[0]])
  p1 = projection_lnglat(p1, epsg_from, epsg_to)
  p2 = projection_lnglat(p2, epsg_from, epsg_to)
  return Point(p1).distance(Point(p2))


def get_geoms(geo: BaseGeometry) -> List:
  """
  将geo中包含的LineString、Point、Polygon取出存放到list中
  Args:
    geo: geometry类型的数据

  Returns:
    List：geo中包含的LineString、Point、Polygon
  """
  list_geo = []
  if geo.is_empty:
    list_geo.append(geo)
  elif hasattr(geo, 'geoms'):
    for g in geo.geoms:
      list_geo.extend(get_geoms(g))
  else:
    list_geo.append(geo)
  return list_geo



import numpy as np
import pandas as pd
from shapely.geometry import Point

from ricco.util.base import is_empty


def test_is_empty():
  assert is_empty(1) is False
  assert is_empty(None) is True
  assert is_empty(np.nan) is True
  assert is_empty([]) is True
  assert is_empty({}) is True
  assert is_empty(pd.DataFrame()) is True
  assert is_empty(Point()) is True



import re

import pandas as pd

from .strings import drop_repeat_string
from .util import relstrip

RES_PRE = {
  '省': None,
  '市': None,
  '区': None,
  '街道': None,
  '居委': None,
  '路': None,
  '弄': None,
  '号': None,
  'subfix': None,
  'extra': None,
  'format': None,
}


def drop_repeat_hao(string: str):
  """将重复的xx号中间的部分全部删掉"""
  ls = re.findall('\d+号', string)
  if len(ls) >= 2:
    _df = pd.DataFrame({'号': ls})
    _df = pd.DataFrame(_df['号'].value_counts()).reset_index(names='name')
    if _df['号'][0] >= 2:
      sep = _df['name'][0]
      _parts = string.split(sep)
      return sep.join([_parts[0], _parts[-1]])
  return string


def split_by_keyword(string):
  """按照关键词对地址进行初步拆分"""
  seps = {
    '省': ['省', '自治区'],
    '市': ['市'],
    '区': ['区', '县'],
    '街道': ['镇', '街道'],
    '居委': ['居委'],
    '路': ['路', '大道'],
    '弄': ['弄'],
  }
  if '集市' in string:
    string = string.replace('集市', 'vvjishivv')
  res = RES_PRE.copy()
  for key, kwds in seps.items():
    res[key] = None
    for kwd in kwds:
      s_list = string.split(kwd, 1)
      if len(s_list) == 2:
        res[key] = s_list[0] + key
        string = s_list[1]
        break
  res['extra'] = None if string == '' else string
  return res


def clean_after_split(string: str):
  """清洗提取后的值"""
  if not isinstance(string, str):
    return string
  # 删除分隔符
  for pat in ['/', '-', '—', '－']:
    string = string.strip(pat)
  string = string.replace('vvjishivv', '集市')
  return None if string == '' else string


def process_by_dict(input_dict: dict):
  """批量处理提取后的值"""
  for k, v in input_dict.items():
    input_dict[k] = clean_after_split(v)
  return input_dict


def is_std_hao(string):
  """
  检查号相关的字符串是否为标准格式，仅检查是否为"xx号、xx号"的形式，不检查重复值和顺序
  """
  ls = string.split('、')
  return all([re.match('^\d+号$', i) for i in ls])


def hao_format(string):
  """对"号"部分进行标准化，主要操作为去重和排序"""
  ls = list(set(string.split('、')))
  ls.sort()
  return '、'.join(ls)


def extract_seps(string):
  _ = []
  ls = re.split('[、|，]', string)
  for i in ls:
    if re.match('^\d+$', i):
      _.append(i + '号')
    if re.match('^\d+号$', i):
      _.append(i)
  _ = list(set(_))
  _.sort()
  return '、'.join(_)


def extra_process(string):
  """提取xx号和后缀，xx号之后的都认为是后缀"""
  res = {
    '号': None,
    'subfix': None,
  }

  if pd.isna(string) or string == '':
    return

  # 类似标准格式的，标准化后直接返回
  if is_std_hao(string):
    res['号'] = hao_format(string)
    return res

  # 路名之后为纯数字的认为是xx号
  if pre_str := re.findall('^\d+$', string):
    res['号'] = pre_str[0] + '号'
    return res

  # 范围号提取，如221-223是221号、222号、223号
  if pre_str := re.findall('^\d+号?[\-—－一]\d+号?', string):
    pre_str = pre_str[0]
    ls = [int(i) for i in re.findall('\d+', pre_str)]
    sub_str = relstrip(string, pre_str).lstrip('号')
    res['subfix'] = None if sub_str == '' else sub_str
    if ls[0] > ls[1]:
      res['号'] = str(ls[0]) + '号'
      res['subfix'] = str(ls[1]) + sub_str
    # elif any([i in string for i in ['奇', '偶', '单', '双']]):
    else:
      res['号'] = '、'.join([f'{i}号' for i in range(ls[0], ls[1] + 2, 2)])
    # 认为同一个地址为xx号-xx号的"号"要么是奇数要么是偶数
    # else:
    #   res['号'] = '、'.join([f'{i}号' for i in range(ls[0], ls[1] + 1, 1)])
    return res

  # 枚举号提取，如"221号，222号"或"221、222号"，提取结果为："221号、222号"
  if pre_str := re.findall('^\d+号?(?:[、，]\d+号?)*', string):
    pre_str = pre_str[0]
    res['号'] = extract_seps(pre_str)
    res['subfix'] = clean_after_split(relstrip(string, pre_str))
    return res

  # 直接提取xx号，仅提取第一个
  if pre_str := re.findall('^\d+号', string):
    pre_str = pre_str[0]
    sub_str = relstrip(string, pre_str).lstrip('号')
    res['号'] = pre_str
    res['subfix'] = clean_after_split(sub_str)
    return res

  if not res['号']:
    res['subfix'] = string

  return res


def extract_extra(res):
  if not res['extra']:
    return res
  res['extra'] = drop_repeat_hao(res['extra'])
  _ = extra_process(res['extra'])
  res['号'] = _['号']
  res['subfix'] = _['subfix']
  return res


def pre_process(string: str):
  """对字符串进行预处理，统一部分字符，删除空格及特殊字符"""
  if pd.isna(string):
    return

  string = str(string)
  for s in ['—', '－', '~', '～']:
    string = string.replace(s, '-')
  for s in ['?', '？']:
    string = string.replace(s, '')
  string = ''.join(string.split())

  # 将"xx甲号"改为"xx号甲"
  if ls := re.findall('\d+[甲乙丙丁戊己庚辛壬癸]号', string):
    for str_o in ls:
      s1 = re.findall('^\d+', str_o)[0]
      s2 = re.findall('[甲乙丙丁戊己庚辛壬癸]', str_o)[0]
      str_d = f'{s1}号{s2}'
      string = string.replace(str_o, str_d)
  if string == '':
    return
  return string


def formated(res):
  """拼接为标准的地址"""
  res['format'] = (
      (res['路'] or '') +
      (res['弄'] or '') +
      (res['号'] or '') +
      ('-' if res['subfix'] and any(
          [res['路'], res['弄'], res['号']]
      ) else '') +
      (res['subfix'] or '')
  )
  return res


def main(string):
  string = pre_process(string)

  if pd.isna(string):
    return RES_PRE

  string = drop_repeat_string(string)
  # 初步提取
  res = split_by_keyword(string)
  # 提取结果简单清洗
  res = process_by_dict(res)
  # 对extra进行再次提取
  res = extract_extra(res)
  # 生成标准列
  res = formated(res)
  return res



import warnings

import pandas as pd

from .base import ensure_list
from .base import not_empty
from .util import and_
from .util import is_digit
from .util import is_unique_series


def skip_column(df: pd.DataFrame, col: str, skip_if_not_exists: bool = True):
  if col not in df:
    if skip_if_not_exists:
      warnings.warn(f'"{col}"列不存在，已跳过')
      return True
    else:
      raise KeyError(f'"{col}"列不存在')
  return False


def assert_columns_exists(df: pd.DataFrame, columns: list):
  """检查列是否存在"""
  columns = ensure_list(columns)
  cols = [c for c in columns if c not in df]
  if cols:
    raise AssertionError(f'缺失"{cols}"列')


def assert_not_empty_str(df: pd.DataFrame,
                         col: str,
                         skip_if_not_exists: bool = True):
  """校验是否存在空白字符串"""
  if skip_column(df, col, skip_if_not_exists):
    return
  if (df[col] == '').any():
    raise AssertionError(f'"{col}"列存在空白字符串')


def assert_not_null(df: pd.DataFrame,
                    col: str,
                    skip_if_not_exists: bool = True):
  """检查是否非空（空白字符串认为是空值）"""
  if skip_column(df, col, skip_if_not_exists):
    return
  if not isinstance(col, str):
    raise TypeError('col参数仅支持str类型')
  if df[col].isna().any():
    raise AssertionError(f'"{col}"列存在空值')
  assert_not_empty_str(df, col)


def assert_values_in(df: pd.DataFrame,
                     col: str,
                     enums: (dict, list),
                     skip_if_not_exists: bool = True):
  """
  检查Dataframe中某一列的值是否在指定的值的范围内
  Args:
    df: 要检查的dataframe
    col: 列名
    enums: 指定的enum值，当传入dict时，包含在key和value中的值都通过
    skip_if_not_exists: 当列不存在时是否跳过
  """
  if skip_column(df, col, skip_if_not_exists):
    return
  if isinstance(enums, dict):
    vs = [*list(enums.keys()), *list(enums.values())]
  elif isinstance(enums, list):
    vs = enums
  else:
    raise TypeError('enums类型错误，list or dict')
  rv = [t for t in df[col].unique().tolist() if not_empty(t) and t not in vs]
  if rv:
    raise AssertionError(
        f'{rv}不是"{col}"列有效的enum值'
    )


def assert_series_unique(df: pd.DataFrame,
                         columns: (str, list) = '名称',
                         text: str = '',
                         ignore_na=False):
  """
  检查并输出重复项
  Args:
    df: 要检查的Dataframe
    columns: 唯一的列
    text: 输出的文案
    ignore_na: 是否忽略空值，默认False，为True时，有空值的行不参与校验
  """
  columns = ensure_list(columns)
  if not is_unique_series(df, columns, ignore_na=ignore_na):
    if ignore_na:
      df = df[and_(*[df[c].notna() for c in columns])]
    df = df[df.duplicated(subset=columns)][columns].sort_values(columns)
    df = df.drop_duplicates()
    ls = df.astype(str).to_dict('records')
    ls = [','.join(list(i.values())) for i in ls]
    info = '\n->'.join(ls)
    raise AssertionError(f'{text}{columns}列中存在重复值:\n->{info}')


def assert_series_digit(df: pd.DataFrame, col: str):
  """检查一列是否可以转为数值型"""
  values = df[col].unique().tolist()
  rv = [i for i in values if not is_digit(i) and not_empty(i)]
  if rv:
    raise AssertionError(f'"{col}"列应为数值型，{rv}无法转换为数值型')



import math

import geojson.utils
from shapely.geometry.base import BaseGeometry
from shapely.geometry.base import BaseMultipartGeometry
from shapely.ops import transform as sh_transform
from tqdm import tqdm

from ..util.base import is_empty
from ..util.decorator import check_null
from .df import _ensure_geometry
from .df import shapely2x
from .util import infer_geom_format

earthR = 6378245.0
x_pi = math.pi * 3000.0 / 180.0


class SRS:
  """空间参考系统(Spatial Reference System)"""
  #: 世界大地测量系统 (World Geodetic System 1984)
  wgs84 = 'wgs84'
  #: 百度坐标系
  bd09 = 'bd09'
  #: 国测局坐标（或火星坐标）
  gcj02 = 'gcj02'


def out_of_china(lat, lng):
  return not (72.004 <= lng <= 137.8347 and 0.8293 <= lat <= 55.8271)


def transform(x, y):
  xy = x * y
  abs_x = math.sqrt(abs(x))
  xpi = x * math.pi
  ypi = y * math.pi
  d = 20.0 * math.sin(6.0 * xpi) + 20.0 * math.sin(2.0 * xpi)

  lat = d
  lng = d

  lat += 20.0 * math.sin(ypi) + 40.0 * math.sin(ypi / 3.0)
  lng += 20.0 * math.sin(xpi) + 40.0 * math.sin(xpi / 3.0)

  lat += 160.0 * math.sin(ypi / 12.0) + 320 * math.sin(ypi / 30.0)
  lng += 150.0 * math.sin(xpi / 12.0) + 300.0 * math.sin(xpi / 30.0)

  lat *= 2.0 / 3.0
  lng *= 2.0 / 3.0

  lat += -100.0 + 2.0 * x + 3.0 * y + 0.2 * y * y + 0.1 * xy + 0.2 * abs_x
  lng += 300.0 + x + 2.0 * y + 0.1 * x * x + 0.1 * xy + 0.1 * abs_x

  return lat, lng


def delta(lat, lng):
  ee = 0.00669342162296594323
  d_lat, d_lng = transform(lng - 105.0, lat - 35.0)
  rad_lat = lat / 180.0 * math.pi
  magic = math.sin(rad_lat)
  magic = 1 - ee * magic * magic
  sqrt_magic = math.sqrt(magic)
  d_lat = (d_lat * 180.0) / (
      (earthR * (1 - ee)) / (magic * sqrt_magic) * math.pi)
  d_lng = (d_lng * 180.0) / (earthR / sqrt_magic * math.cos(rad_lat) * math.pi)
  return d_lat, d_lng


def wgs2gcj(wgs_lat, wgs_lng):
  if out_of_china(wgs_lat, wgs_lng):
    return wgs_lat, wgs_lng
  dlat, dlng = delta(wgs_lat, wgs_lng)
  return wgs_lat + dlat, wgs_lng + dlng


def gcj2wgs(gcj_lat, gcj_lng):
  if out_of_china(gcj_lat, gcj_lng):
    return gcj_lat, gcj_lng
  dlat, dlng = delta(gcj_lat, gcj_lng)
  return gcj_lat - dlat, gcj_lng - dlng


def gcj2bd(gcj_lat, gcj_lng):
  if out_of_china(gcj_lat, gcj_lng):
    return gcj_lat, gcj_lng
  x = gcj_lng
  y = gcj_lat
  z = math.hypot(x, y) + 0.00002 * math.sin(y * x_pi)
  theta = math.atan2(y, x) + 0.000003 * math.cos(x * x_pi)
  bd_lng = z * math.cos(theta) + 0.0065
  bd_lat = z * math.sin(theta) + 0.006
  return bd_lat, bd_lng


def bd2gcj(bd_lat, bd_lng):
  if out_of_china(bd_lat, bd_lng):
    return bd_lat, bd_lng
  x = bd_lng - 0.0065
  y = bd_lat - 0.006
  z = math.hypot(x, y) - 0.00002 * math.sin(y * x_pi)
  theta = math.atan2(y, x) - 0.000003 * math.cos(x * x_pi)
  gcj_lng = z * math.cos(theta)
  gcj_lat = z * math.sin(theta)
  return gcj_lat, gcj_lng


def wgs2bd(wgs_lat, wgs_lng):
  return gcj2bd(*wgs2gcj(wgs_lat, wgs_lng))


def bd2wgs(bd_lat, bd_lng):
  return gcj2wgs(*bd2gcj(bd_lat, bd_lng))


_fn_mapping = {
  (SRS.bd09, SRS.wgs84): bd2wgs,
  (SRS.gcj02, SRS.wgs84): gcj2wgs,
  (SRS.wgs84, SRS.bd09): wgs2bd,
  (SRS.gcj02, SRS.bd09): gcj2bd,
  (SRS.wgs84, SRS.gcj02): wgs2gcj,
  (SRS.bd09, SRS.gcj02): bd2gcj,
}


def _coord_transform(lng: float, lat: float, from_srs: (SRS, str),
                     to_srs: (SRS, str)):
  """
  坐标系转换
  Args:
    lng: 输入的经度
    lat: 输入的纬度
    from_srs: 输入坐标的格式
    to_srs: 输出坐标的格式
  """
  if from_srs == to_srs:
    return lng, lat

  if is_empty(lng) or is_empty(lat):
    return None, None

  key = (from_srs, to_srs)
  if key not in _fn_mapping:
    raise NotImplementedError(
        'not support transformation from %s to %s' % (from_srs, to_srs))
  lat, lng = _fn_mapping[key](lat, lng)
  return lng, lat


def coord_transform_geojson(obj: dict, from_srs: SRS, to_srs: SRS):
  """对GeoJSON格式内的所有坐标点执行坐标系转换

  Examples:
    point_bd09 = Point([1, 2])
    geojson = mapping(point_bd09)
    transformed = coord_transform(geojson, SRS.bd09, SRS.wgs84)
    point_wgs84 = shape(transformed)
  """
  return geojson.utils.map_tuples(
      lambda c: _coord_transform(c[0], c[1], from_srs, to_srs), obj)


@check_null()
def _coord_transform_geometry(geo: (BaseGeometry, BaseMultipartGeometry),
                              from_srs: SRS,
                              to_srs: SRS):
  """对Geomery内的所有点进行坐标转换，返回转换后的Geometry

  该方法可以支持所有的Shapely Geometry形状，包括Point, Line, Polygon,
  MultiPloygon等，返回的Geometry和输入的形状保持一致
  Args:
    geo: 输入的shapely Geometry
    from_srs: 输入的坐标格式
    to_srs: 输出的坐标格式
  Returns:
    转换后的shapely Geometry
  """
  return sh_transform(
      lambda x, y, z=None: _coord_transform(x, y, from_srs, to_srs), geo)


def coord_trans_x2y(df,
                    srs_from: (SRS, str),
                    srs_to: (SRS, str),
                    c_lng: str = 'lng',
                    c_lat: str = 'lat'):
  """
  经纬度类型坐标坐标批量转换工具
  Args:
    df: 输入的dataframe，必须要有geometry列
    srs_from: 当前坐标系，可选wgs84,bd09,gcj02
    srs_to: 要转的坐标系，可选wgs84,bd09,gcj02
    c_lng: 经度列名
    c_lat: 纬度列名
  """
  df = df.copy()
  tqdm.pandas(desc=f'{srs_from}->{srs_to}')
  df[[c_lng, c_lat]] = df.progress_apply(
      lambda r: _coord_transform(r[c_lng], r[c_lat], srs_from, srs_to),
      axis=1, result_type='expand'
  )
  return df


def coord_trans_geom(df,
                     srs_from: (SRS, str),
                     srs_to: (SRS, str),
                     c_geometry: str = 'geometry',
                     geometry_format=None):
  """
  geometry类型坐标批量转换工具
  Args:
    df: 输入的dataframe，必须要有geometry列
    srs_from: 当前坐标系，可选wgs84,bd09,gcj02
    srs_to: 要转的坐标系，可选wgs84,bd09,gcj02
    c_geometry: 要转换的geometry列名
    geometry_format: 指定要输出的geometry格式，默认返回和原来相同的geometry格式
  """

  assert c_geometry in df
  df = df.copy()
  if not geometry_format:
    geometry_format = infer_geom_format(df[c_geometry])
  df_temp = _ensure_geometry(df)
  tqdm.pandas(desc=f'{srs_from}->{srs_to}')
  df_temp[c_geometry] = df_temp[c_geometry].progress_apply(
      lambda x: _coord_transform_geometry(x, srs_from, srs_to)
  )
  del df[c_geometry]
  df = df.join(df_temp[[c_geometry]], how='left')
  return shapely2x(df, geometry_format=geometry_format, geometry=c_geometry)


def coord_transformer(df,
                      srs_from: (SRS, str),
                      srs_to: (SRS, str),
                      c_lng: str = 'lng',
                      c_lat: str = 'lat',
                      c_geometry: str = 'geometry',
                      geometry_format=None):
  """
  坐标转换工具，优先转geometry列
  Args:
    df: 输入的dataframe，必须要有geometry列
    srs_from: 当前坐标系，可选wgs84,bd09,gcj02
    srs_to: 要转的坐标系，可选wgs84,bd09,gcj02
    c_lng: 经度列名
    c_lat: 纬度列名
    c_geometry: 要转换的geometry列名
    geometry_format: 指定要输出的geometry格式，默认返回和原来相同的geometry格式
  """
  if c_geometry in df:
    return coord_trans_geom(df, srs_from, srs_to, c_geometry, geometry_format)
  elif c_lat in df and c_lng in df:
    return coord_trans_x2y(df, srs_from, srs_to, c_lng, c_lat)
  else:
    raise KeyError(f'文件中必须有经纬度列或"{c_geometry}"列')



import logging
import warnings

import pandas as pd
from shapely.geometry.base import BaseGeometry


def ensure_list(val):
  """将标量值和Collection类型都统一转换为LIST类型"""
  if val is None:
    return []
  if isinstance(val, list):
    return val
  if isinstance(val, (set, tuple)):
    return list(val)
  return [val]


def ensure_ext(ext: str):
  if isinstance(ext, str) and ext != '' and not ext.startswith('.'):
    return '.' + ext
  return ext


def is_empty(x) -> bool:
  """
  判断是否为空值，以下值认为是空白
    - 空白列表、字典, 如：[], {}，
    - 空白Dataframe、series, 如：pd.DataFrame()
    - 空白shapely格式的geometry，如：Point(np.nan, np.nan)
  """
  if isinstance(x, (list, dict, tuple)):
    return False if x else True
  if isinstance(x, (pd.DataFrame, pd.Series)):
    return x.empty
  if isinstance(x, BaseGeometry):
    return x.is_empty
  return pd.isna(x)


def not_empty(x) -> bool:
  """判断是否非空"""
  return not is_empty(x)


def second_to_dhms(second):
  """将秒转为天、时、分、秒"""
  d = int(second // 86400)
  second = second % 86400
  h = int(second // 3600)
  second = second % 3600
  m = int(second // 60)
  second = second % 60
  if d + h + m == 0:
    second = round(second, 2)
  else:
    second = int(second)
  return d, h, m, second


def second_to_desc(second):
  """将秒转为时间描述"""
  d, h, m, second = second_to_dhms(second)
  if d + h + m == 0:
    return f'{second:.2f}s'
  if d + h == 0:
    return f'{m}m {second:.0f}s'
  if d == 0:
    return f'{h}h {m}m {second:.0f}s'
  return f'{d}d {h}h {m}m {second:.0f}s'


def warn_(msg, if_or_not=True, mode='warning'):
  if if_or_not:
    if mode == 'warning':
      warnings.warn(msg)
    if mode == 'logging':
      logging.warning(msg)



import warnings

from ..geometry.coord_trans import gcj2bd
from ..geometry.coord_trans import gcj2wgs
from ..util.decorator import check_null

DEFAULT_RES = {
  'rv': None, 'score': -1,
  'lng': None, 'lat': None,
  'source': None
}


class MapUrls:
  # 百度地理编码
  baidu = 'http://api.map.baidu.com/geocoding/v3/'
  # 百度地点检索
  baidu_poi = 'http://api.map.baidu.com/place/v2/search'
  # 高德地理编码
  amap = 'https://restapi.amap.com/v3/geocode/geo'
  # 高德地点检索
  amap_poi = 'https://restapi.amap.com/v3/place/text'
  # MC geocode服务
  mdt = 'https://geocode.idatatlas.com/geocode'


class MapKeys:
  amap = '7c14855824549a84c543e48990239f3d'
  baidu = '9Fy1lMHbwpr07WVBFPLw9vpntGUSOUMN'


def gcj2xx(lnglat, srs):
  assert srs in ('wgs84', 'bd09', 'gcj02'), '可选参数为bd09,wgs84,gcj02'
  if srs == 'wgs84':
    return gcj2wgs(float(lnglat[1]), float(lnglat[0]))
  elif srs == 'bd09':
    return gcj2bd(float(lnglat[1]), float(lnglat[0]))
  else:
    return float(lnglat[1]), float(lnglat[0])


def error_baidu(js):
  if js['status'] != 0:
    warnings.warn(
        f'接口错误，状态码【{js["status"]}】，错误原因请查阅：https://lbsyun.baidu.com/index.php?title=webapi/appendix')


def error_amap(js):
  if js['status'] != '1':
    warnings.warn(
        f'接口错误，状态码【{js["infocode"]}】，错误原因请查阅：https://lbs.amap.com/api/webservice/guide/tools/info')


@check_null(default_rv='')
def fix_address(string):
  string = str(string)
  for _ in ['&', '%', '#', '@', '$', '|']:
    string = string.replace(_, '')
  return string


@check_null(default_rv='中国')
def fix_city(string):
  return str(string).rstrip('市')


def rv_score(city, address, rv):
  from fuzzywuzzy import fuzz
  if not all([address, rv]):
    return -1
  city = fix_address(city)
  return max(
      fuzz.ratio(rv.lstrip(city), address.lstrip(city)),
      fuzz.partial_ratio(rv.lstrip(city), address.lstrip(city))
  )



from ricco.util.strings import drop_repeat_string
from ricco.util.strings import get_breaks
from ricco.util.strings import get_city_and_region
from ricco.util.strings import get_single_list


def test_drop_repeat_string():
  assert drop_repeat_string(
      '112341234'
  ) == '11234'
  assert drop_repeat_string(
      '11'
  ) == '11'
  assert drop_repeat_string(
      '123123123123'
  ) == '123'
  assert drop_repeat_string(
      '上海市/普陀区/甘泉路街道/志丹路志丹路490号/志丹路志丹路490号'
  ) == '上海市/普陀区/甘泉路街道/志丹路490号'


def test_get_single_list():
  assert get_single_list(0, 7, 3) == [[0, 3], [3, 6], [6, 7]]


def test_get_breaks():
  assert get_breaks(10, 3) == [
    [[0, 3], [3, 6], [6, 9], [9, 10]],
    [[0, 1], [1, 4], [4, 7], [7, 10]],
    [[0, 2], [2, 5], [5, 8], [8, 10]]
  ]


def test_extract_city():
  s = '上海市杨浦区政立路111号'
  assert get_city_and_region(s) == ('上海市', '杨浦区', ['上海市'], ['杨浦区'])

  s = '浙江省杭州市滨江区西兴街道泰安路199号'
  assert get_city_and_region(s) == ('杭州市', '滨江区', ['杭州市'], ['滨江区'])

  s = '云南省红河哈尼族彝族自治州个旧市大屯镇团山村万城米兰春天25幢2号'
  assert get_city_and_region(s) == (
    '红河哈尼族彝族自治州', '个旧市', ['红河哈尼族彝族自治州'], ['个旧市']
  )

  s = '内蒙古乌兰察布市察哈尔右翼前旗平地泉北方汽车城B厅35号'
  assert get_city_and_region(s) == (
    '乌兰察布市', '察哈尔右翼前旗', ['乌兰察布市'], ['察哈尔右翼前旗']
  )

  s = '甘肃省武威市天祝藏族自治县东大滩乡圈湾村'
  assert get_city_and_region(s) == (
    '武威市', '天祝藏族自治县', ['武威市'], ['天祝藏族自治县']
  )

  s = '内蒙古自治区锡林郭勒盟锡林浩特市污水处理厂附近工业园区17号'
  assert get_city_and_region(s) == (
    '锡林郭勒盟', '锡林浩特市', ['锡林郭勒盟'], ['锡林浩特市']
  )

  s = '河北省石家庄市桥西区新石中路168号15栋2单元402室'
  assert get_city_and_region(s) == (
    '石家庄市', '桥西区', ['石家庄市'], ['桥西区']
  )



import warnings

import requests

from ..util.decorator import check_null
from .util import DEFAULT_RES
from .util import MapKeys
from .util import MapUrls
from .util import error_amap
from .util import fix_address
from .util import fix_city
from .util import gcj2xx
from .util import rv_score

KEY = MapKeys.amap


@check_null()
def address_json(address: str, city: str, key=None):
  """高德地理编码接口"""
  if not key:
    key = KEY
  url = f'{MapUrls.amap}?address={address}&city={city}&key={key}'
  js = requests.get(url).json()
  error_amap(js)
  if js['status'] == '1' and int(js['count']) >= 1:
    return js['geocodes'][0]


@check_null()
def place_json(address: str, city: str, key=None):
  """高德地点检索接口"""
  if not key:
    key = KEY
  url = f'{MapUrls.amap_poi}?keywords={address}&city={city}&key={key}'
  js = requests.get(url).json()
  error_amap(js)
  if js['status'] == '1' and int(js['count']) >= 1:
    return js['pois'][0]


@check_null()
def get_amap(*,
             address,
             city,
             source,
             with_detail=True,
             disable_cache=False,
             key=None):
  """MC geocode服务"""
  assert source in ('amap', 'amap_poi')
  url = f'{MapUrls.mdt}?address={address}&city={city}&disable_cache={disable_cache}&with_detail={with_detail}&source={source}'
  req = requests.get(url)
  if req.status_code == 200:
    try:
      return req.json()['result'][0]['extra']
    except Exception as e:
      warnings.warn(f'{e}，{req}')
      return
  if req.status_code in (400, 403):
    if source == 'amap':
      return address_json(address=address, city=city, key=key)
    if source == 'amap_poi':
      return place_json(address=address, city=city, key=key)
  else:
    warnings.warn(f'Unexpected status_code：{req.status_code}，{city}|{address}')


@check_null(default_rv=DEFAULT_RES)
def get_address_amap(address: str, city: str,
                     srs: str = 'wgs84', key=None) -> dict:
  result = DEFAULT_RES.copy()
  source = 'amap'
  city = fix_city(city)
  address = fix_address(address)
  address_dict = get_amap(address=address, city=city, source=source, key=key)
  if address_dict:
    result['rv'] = address_dict['formatted_address']
    latlng = gcj2xx(address_dict['location'].split(','), srs=srs)
    result['lng'] = latlng[1]
    result['lat'] = latlng[0]
    result['score'] = rv_score(city, address, address_dict['formatted_address'])
    result['source'] = source
  return result


@check_null(default_rv=DEFAULT_RES)
def get_place_amap(address: str, city: str,
                   srs: str = 'wgs84', key=None) -> dict:
  result = DEFAULT_RES.copy()
  source = 'amap_poi'
  city = fix_city(city)
  address = fix_address(address)
  poi_dict = get_amap(address=address, city=city, source='amap_poi', key=key)
  if poi_dict:
    result['rv'] = poi_dict['name']
    latlng = gcj2xx(poi_dict['location'].split(','), srs=srs)
    result['lng'] = latlng[1]
    result['lat'] = latlng[0]
    result['score'] = rv_score(city, address, poi_dict['name'])
    result['source'] = source
  return result



import numpy as np

from ricco.util.util import eval_
from ricco.util.util import extract_num
from ricco.util.util import fix_str
from ricco.util.util import get_shortest_element
from ricco.util.util import house_type_format
from ricco.util.util import interchange_dict
from ricco.util.util import is_digit
from ricco.util.util import list2dict
from ricco.util.util import pinyin
from ricco.util.util import relstrip
from ricco.util.util import rerstrip
from ricco.util.util import rstrip_d0
from ricco.util.util import segment
from ricco.util.util import sort_by_list
from ricco.util.util import to_bool
from ricco.util.util import to_float
from ricco.util.util import union_list_v2
from ricco.util.util import union_str_v2
from ricco.util.util import remove_null_in_dict


def test_relstrip():
  assert relstrip('257号2234', '257号') == '2234'


def test_rerstrip():
  assert rerstrip('257号2234', '234') == '257号2'


def test_get_get_shortest_element():
  assert get_shortest_element(
      ['123', '1', '3211', 'abcdef']
  ) == '1'
  assert get_shortest_element(
      ['123', 1, '3211', 'abcdef']
  ) == 1
  assert get_shortest_element(
      [None, 'sssssss', 'sssssss', '3211', 'abcdef']
  ) == '3211'


def test_pinyin():
  assert pinyin('测试') == 'ceshi'
  assert pinyin('test') == 'test'


def test_extract_num():
  string = 'fo13--;gr35.3'
  assert extract_num(string) == ['13', '35.3']
  assert extract_num(string, num_type='int') == [13, 35]
  assert extract_num(string, num_type='float') == [13.0, 35.3]


def test_to_float():
  assert to_float('6.5') == 6.5
  assert to_float('6.5%') == 0.065
  assert to_float('6.5%', ignore_pct=True) == 6.5
  assert to_float('p6.5%') == 0.065
  assert to_float('p6.5') == 6.5
  assert to_float('p6.5xx3.5', multi_warning=False) == 5.0
  assert to_float('p6.5xx3.5', rex_method='min', multi_warning=False) == 3.5
  assert to_float('p6.5xx3.5', rex_method='max', multi_warning=False) == 6.5
  assert to_float('p6.5xx3.5', rex_method='sum', multi_warning=False) == 10.0


def test_segment():
  assert segment(55, 20) == '40-60'
  assert segment(55, 20, sep='--', unit='米') == '40--60米'
  assert segment(55, [20]) == '20以上'
  assert segment(10, [20, 50], unit='米') == '20米以下'
  assert segment(20, [20, 50], unit='米') == '20-50米'
  assert segment(50, [20, 50], unit='米') == '50米以上'


def test_house_type_format():
  assert house_type_format('一室一厅') == '1房'
  assert house_type_format('两室一厅') == '2房'
  assert house_type_format('3室两厅') == '3房'
  assert house_type_format('4房1厅') == '4房'
  assert house_type_format('5室1厅') == '5房及以上'
  assert house_type_format('8室1厅') == '5房及以上'


def test_list2dict():
  assert list2dict([1, 2, 3]) == {0: 1, 1: 2, 2: 3}


def test_eval():
  assert eval_(None) == None
  assert eval_('[1]') == [1]
  assert eval_("{'a':1}") == {'a': 1}


def test_union_str():
  assert union_str_v2('abc', 'def') == 'abcdef'
  assert union_str_v2('abc', 'def', sep='|') == 'abc|def'
  assert union_str_v2('abc', None) == 'abc'


def test_union_list():
  assert union_list_v2([1, 2, 3]) == [1, 2, 3]
  assert union_list_v2([1, 2, 3], [4, 5]) == [1, 2, 3, 4, 5]
  assert union_list_v2([1, 2, 3], 4, 5) == [1, 2, 3, 4, 5]


def test_sort_by_list():
  src_list = [1, 2, 3, 4, 5]
  by_list = [4, 5, 3, 6]
  assert sort_by_list(src_list, by_list, filter_=False) == [4, 5, 3, 1, 2]
  assert sort_by_list(src_list, by_list, filter_=True) == [4, 5, 3]


def test_fix_empty_str():
  assert fix_str(1) == 1
  assert fix_str('a') == 'a'
  assert fix_str(' a ') == 'a'
  assert fix_str('\na\n') == 'a'
  assert fix_str('') == None
  assert fix_str(' ') == None
  assert fix_str(' \n') == None
  assert fix_str(' ') == None


def test_is_digit():
  assert is_digit(1) is True
  assert is_digit('1') is True
  assert is_digit('1.1') is True
  assert is_digit('a') is False
  assert is_digit(None) is False
  assert is_digit(np.nan) is False


def test_rstrip_d0():
  assert rstrip_d0('') == ''
  assert rstrip_d0(0) == 0
  assert rstrip_d0(1.0) == '1'
  assert rstrip_d0('a') == 'a'
  assert not rstrip_d0(None)
  assert not rstrip_d0(np.nan)


def test_to_bool():
  assert to_bool(1) is True
  assert to_bool('是') is True
  assert to_bool('0') is False
  assert to_bool('否') is False
  assert to_bool(None) is False
  assert to_bool(None, na=True) is True
  assert to_bool(123) is False
  assert to_bool(123, other=True) is True
  assert to_bool(123, other='abc') == 'abc'


def test_interchange_dict():
  assert interchange_dict({1: 2}) == {2: 1}


def test_remove_null_in_dict():
  assert remove_null_in_dict({1: 2, 3: None, 4: None}) == {1: 2}



__version__ = '1.3.0'

from .etl import file
from .etl import load
from .etl import transformer
from .etl.extract import rdf
from .etl.extract import rdf_by_dir
from .etl.file import file_to_x
from .etl.file import reshape_files
from .etl.file import split2x
from .etl.load import to_csv_by_line
from .etl.load import to_file
from .etl.transformer import expand_dict
from .etl.transformer import is_changed
from .etl.transformer import keep_best_unique
from .etl.transformer import split_to_rows
from .etl.transformer import table2dict
from .etl.transformer import update_df
from .geocode.geocode import geocode
from .geocode.geocode import geocode_df
from .geocode.geocode import geocode_v2
from .geometry import coord_trans
from .geometry import df as geom
from .geometry.coord_trans import coord_trans_geom
from .geometry.coord_trans import coord_trans_x2y
from .geometry.coord_trans import coord_transformer
from .geometry.df import auto2shapely
from .geometry.df import auto2x
from .geometry.df import get_area
from .geometry.df import mark_tags_v2
from .geometry.df import nearest_neighbor
from .geometry.df import shapely2x
from .geometry.util import wkb_dumps
from .geometry.util import wkb_loads
from .geometry.util import wkt_dumps
from .geometry.util import wkt_loads
from .local import Rc
from .util import dt
from .util import id_number
from .util import os
from .util import phone_number
from .util import strings
from .util.assertion import assert_columns_exists
from .util.assertion import assert_not_empty_str
from .util.assertion import assert_not_null
from .util.assertion import assert_series_digit
from .util.assertion import assert_series_unique
from .util.assertion import assert_values_in
from .util.base import ensure_list
from .util.base import is_empty
from .util.base import not_empty
from .util.district import District
from .util.dt import DT
from .util.dt import auto2date
from .util.dt import excel2date
from .util.dt import is_valid_date
from .util.id_number import IDNumber
from .util.os import dir_iter
from .util.os import dir_iter_list
from .util.os import ensure_dirpath_exist
from .util.os import ext
from .util.os import extension
from .util.os import fn
from .util.os import path_name
from .util.os import split_path
from .util.phone_number import PhoneNumber
from .util.strings import drop_repeat_string
from .util.strings import get_city_and_region
from .util.util import and_
from .util.util import ensure_list
from .util.util import extract_num
from .util.util import fix_str
from .util.util import is_empty
from .util.util import list2dict
from .util.util import not_empty
from .util.util import or_
from .util.util import pinyin
from .util.util import re_fast
from .util.util import rstrip_d0
from .util.util import segment
from .util.util import union_list_v2
from .util.util import union_str_v2



from ricco.util.os import extension
from ricco.util.os import path_name


def test_path_name():
  assert path_name('test_util.py') == 'test_util'


def test_extension():
  assert extension('test_util.py') == '.py'



import random
import re
import warnings

from ..resource.patterns import Pattern
from ..resource.phone_number import service_providers


class PhoneNumber:
  def __init__(self, phone_number: str):
    self.phone_number = self.format_phone_number(phone_number)
    self.__3 = self.phone_number[:3]

  @property
  def value(self):
    return self.phone_number

  @staticmethod
  def is_valid(id_number) -> bool:
    """判断手机号是否有效"""
    string = str(id_number)
    if re.match(Pattern.phone_number, string):
      return True
    return False

  @staticmethod
  def not_valid(id_number) -> bool:
    """判断手机号是否无效"""
    return not PhoneNumber.is_valid(id_number)

  def format_phone_number(self, id_number) -> str:
    """对手机号进行校验并转为字符串格式"""
    if isinstance(id_number, (float, int)):
      id_number = str(int(id_number))
    if self.is_valid(id_number):
      return id_number
    else:
      raise ValueError('手机号格式错误')

  @property
  def service_providers(self):
    """手机运营商"""
    if p := service_providers.get(self.__3):
      return p
    else:
      return '未知'

  @property
  def home(self):
    # TODO(wangyukang): 补充手机号归属地
    warnings.warn('方法待补充')
    return

  @classmethod
  def random(cls):
    from ..resource.phone_number import service_providers
    _3 = random.choice(list(service_providers.keys()))
    _8 = random.randint(0, 99999999)
    _8 = str(_8).zfill(8)
    return f'{_3}{_8}'



from .etl.extract import rdf

BASE_PATH = '/Users/ricco/common_data'


class Ricco:
  """通用数据"""

  def aoi(self, city='上海'):
    return rdf(f'{BASE_PATH}/aoi/{city}_全要素底图.csv')

  def region(self, city='上海'):
    return rdf(f'{BASE_PATH}/bd_region/{city}边界_区县最新.csv')

  def street(self, city='上海'):
    return rdf(f'{BASE_PATH}/bd_region/{city}边界_街镇最新.csv')

  def juwei(self, city='上海'):
    return rdf(f'{BASE_PATH}/aoi/{city}_居委边界.csv')

  def xiaoqu(self, city='上海'):
    df = rdf(f'{BASE_PATH}/aoi/{city}边界_AOI.csv')
    return df

  def building_geom(
      self,
      city: str,
      *,
      region: str = None,
      street: str = None,
      subfix: str = '百度',
      all_cols: bool = False,
  ):
    cols = ['oid', '城市', '区县', '街镇', 'geometry']
    df = rdf(f'{BASE_PATH}/aoi/{city}边界_建筑边界_{subfix}.parquet',
             columns=None if all_cols else cols)
    if street:
      return df[df['街镇'] == street]
    if region:
      return df[df['区县'] == region]
    return df


Rc = Ricco()



from shapely.errors import ShapelyDeprecationWarning
from tqdm import tqdm

import geopandas as gpd
import pandas as pd
from shapely.geometry import GeometryCollection
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon
from shapely.ops import unary_union

from shapely.validation import make_valid

import warnings
warnings.filterwarnings('ignore', category=ShapelyDeprecationWarning)

_desc = '''
  拓扑检查
  可检查: 重叠问题, 自相交问题, geometry缺失问题
  可修复: 重叠问题, 自相交问题
'''


def topology_check(gdf: gpd.GeoDataFrame,
                   geo_col='geometry'):
  """
  判断整列geometry是否规范, 仅支持面数据
  检查geometry是否有空值
  检查每个geometr是否存在拓扑问题
  检查整列geometry之间是否存在拓扑问题
  Args:
    gdf: 需要进行拓扑修复的GeoDataFrame
    geo_col: geometry列列名

  Returns:
    geometry数据是否规范
  """
  geo_data = gdf[geo_col]
  if len(geo_data[geo_data.isnull()]):
    raise ValueError('存在空数据')
  if not all([isinstance(geo, (Polygon, MultiPolygon)) for geo in geo_data]):
    raise ValueError('存在非面类型的地理数据')
  for index, geo in tqdm(geo_data.items(),
                         desc='topology_check',
                         total=len(geo_data)):
    if not geo.is_valid:
      return False
    for i, g in geo_data.items():
      if i == index:
        continue
      if geo.intersects(g):
        return False
  return True


def fix_polygon_topology(gdf: gpd.GeoDataFrame,
                         geo_col='geometry',
                         fill_intersects=False,
                         keep_contains=False) -> gpd.GeoDataFrame:
  """
  修复面地理数据的拓扑问题,index不能重复
  Args:
    gdf: 需要进行拓扑修复的GeoDataFrame
    geo_col: geometry列列名
    fill_intersects: 是否填满相交的区域，如果为True的话，两个面相交的区域会随机分配到其中一个面内
    keep_contains: fill_intersects为True时才有效，即是否强制保留被包含的面

  Returns:
    修复完拓扑问题的GeoDataFrame
  """
  if topology_check(gdf, geo_col):
    return gdf
  gdf[geo_col] = series_geometry_fix_topology(gdf[geo_col],
                                              fill_intersects,
                                              keep_contains)
  return gpd.GeoDataFrame(gdf, crs=4326, geometry=geo_col)


def series_geometry_fix_topology(series: pd.Series,
                                 fill_intersects=True,
                                 keep_contains=True):
  """
  Args:
    series: series的index不能重复, 如果keep_contains=True,geometry不能相互包含
    fill_intersects: 是否填满相交的区域，如果为True的话，两个面相交的区域会分配到更靠后的面内
    keep_contains: fill_intersects为True时才有效，即是否强制保留被包含的面

  Returns:

  """
  if any(series.index.duplicated()):
    raise ValueError('duplicated index')

  def fix_valid(x):
    if not x.is_valid:
      valid_res = make_valid(x)
      if isinstance(valid_res, GeometryCollection):
        return unary_union([i for i in valid_res.geoms if isinstance(i, (Polygon, MultiPolygon))])
      return valid_res
    return x

  series = series.apply(fix_valid)

  def ser_intersect(geo_ser: series):
    intersect_info = dict()
    for index, geo in tqdm(geo_ser.items(),
                           desc='get_intersects_info',
                           total=len(geo_ser)):
      intersect_list = []
      intersect_polygon = MultiPolygon()
      for i, g in geo_ser.items():
        if i == index:
          continue
        # TODO(fanjianbang) rtree优化
        if geo.intersects(g):
          intersect_list.append(i)
          intersect_polygon = intersect_polygon.union(g)
      intersect_info[index] = {}
      intersect_info[index]['intersect_list'] = intersect_list
      intersect_info[index]['is_contains'] = intersect_polygon.contains(geo)
    return intersect_info

  def fix_intersects(geo_ser: series):
    intersect_info = ser_intersect(geo_ser)
    res_geo = []
    res_index = []
    contains_geo = []
    contains_index = []
    for index, geo in tqdm(geo_ser.items(),
                           desc='fix_intersects',
                           total=len(geo_ser)):
      info = intersect_info.get(index)
      intersect_list = info.get('intersect_list')
      if info.get('is_contains') and fill_intersects and keep_contains:
        contains_geo.append(geo)
        contains_index.append(index)
      else:
        for i in intersect_list:
          if fill_intersects and i in res_index:
            continue
          geo = geo.difference(geo_ser.loc[i].buffer(1e-7))
        if isinstance(geo, GeometryCollection):
          geo = unary_union([i for i in geo.geoms if isinstance(i, (Polygon, MultiPolygon))])
        res_geo.append(geo)
        res_index.append(index)
    intersect_res = pd.Series(res_geo, index=res_index)
    if contains_index:
      if set(contains_index) == set(geo_ser.index):
        raise ValueError('mutual inclusion geometry')
      contains_res = fix_intersects(pd.Series(contains_geo,
                                              index=contains_index))
      return pd.concat([intersect_res, contains_res])
    else:
      return intersect_res
  res = fix_intersects(series)
  return res



import csv
import json
import logging
import sys

import geopandas as gpd
import pandas as pd
from tqdm import tqdm

from ..geometry.util import wkt_loads
from ..util.base import ensure_list
from ..util.exception import UnknownFileTypeError
from ..util.os import dir_iter
from ..util.os import extension
from ..util.os import path_name


def rdf(
    file_path: str,
    *,
    sheet_name=0,
    sheet_contains: str = None,
    encoding: str = None,
    info: bool = False,
    dtype=None,
    columns: (list, str) = None,
    nrows: int = None,
) -> pd.DataFrame:
  """
  常用文件读取函数，支持
  .csv/.xlsx/.xls/.shp/.parquet/.pickle/.feather/.kml/.ovkml/
  .geojson/.shp/.json等

  Args:
    file_path: 文件路径
    sheet_name: 数据所在sheet的名称，仅对.xlsx/.xls生效
    sheet_contains: 筛选sheet名称中包含此字符串的sheet，仅对.xlsx/.xls生效
    encoding: 编码
    info: 是否打印数据集情况（shape & columns）
    dtype: 指定读取列的类型
    columns: 指定读取的列名
    nrows: 指定读取的行数
  """
  if columns:
    columns = ensure_list(columns)
  ex = extension(file_path)
  if ex == '.csv':
    df = read_csv(
        file_path, dtype=dtype, columns=columns, nrows=nrows, encoding=encoding)
  elif ex in ('.xls', '.xlsx'):
    df = rdxls(
        file_path, sheet_name=sheet_name, sheet_contains=sheet_contains,
        dtype=dtype, columns=columns, nrows=nrows)
  elif ex in ('.shp', '.dbf', '.shx', '.geojson'):
    df = read_shapefile(file_path, encoding=encoding, nrows=nrows)
  elif ex == '.parquet':
    df = pd.read_parquet(file_path, columns=columns)
  elif ex == '.feather':
    df = pd.read_feather(file_path, columns=columns)
  elif ex == '.pickle':
    df = pd.read_pickle(file_path)
  elif ex in ('.kml', '.ovkml'):
    df = read_kml(file_path)
  elif ex == '.json':
    df = read_line_json(file_path, encoding=encoding)
  else:
    raise UnknownFileTypeError('未知的文件扩展名')
  df = df[columns] if columns else df
  df = df.head(nrows) if nrows else df
  if info:
    print(f'shape: {df.shape}')
    print(f'columns: {df.columns.tolist()}')
  return df


def max_grid():
  """防止单个单元格文件过大而报错"""
  maxint = sys.maxsize
  decrement = True
  while decrement:
    decrement = False
    try:
      csv.field_size_limit(maxint)
    except OverflowError:
      maxint = int(maxint / 10)
      decrement = True


def rdxls(
    file_path,
    *,
    sheet_name=0,
    sheet_contains: str = None,
    dtype=None,
    columns: (list, str) = None,
    nrows: int = None
) -> pd.DataFrame:
  """
  读取excel文件
  Args:
    file_path: 文件名
    sheet_name: sheet表的名称
    sheet_contains: sheet表包含的字符串
    dtype: 指定读取列的类型
    columns: 指定读取的列
    nrows: 指定读取的行数
  """
  if sheet_contains:
    assert not sheet_name, 'sheet_name和sheet_contains不能同时指定'
    dataset = pd.read_excel(file_path, sheet_name=None, dtype=dtype,
                            usecols=columns, nrows=1)
    sheets = [i for i in dataset if sheet_contains in i]
    assert len(sheets) == 1, f'0个或多个sheet包含{sheet_contains}'
    sheet_name = sheets[0]
    print(f'通关关键字"{sheet_contains}"匹配到唯一的sheet_name"{sheet_name}"')
  return pd.read_excel(
      file_path, sheet_name=sheet_name, dtype=dtype, usecols=columns,
      nrows=nrows)


def read_csv(file_path: str,
             *,
             encoding: str = None,
             dtype=None,
             columns: (list, str) = None,
             nrows: int = None):
  max_grid()
  encodings = [encoding] if encoding else ['utf-8', 'gbk', 'utf-8-sig']
  for encode in encodings:
    try:
      return pd.read_csv(
          file_path, engine='python', dtype=dtype, usecols=columns,
          nrows=nrows, encoding=encode)
    except UnicodeDecodeError:
      pass
  raise Exception(f'使用encoding{encodings}无法读取文件，请指定')


def read_line_json(file_path, encoding='utf-8') -> pd.DataFrame:
  """
  逐行读取json格式的文件
  Args:
    file_path: 文件名
    encoding: 文件编码，默认为utf-8
  """
  records = []
  with open(file_path, 'r', encoding=encoding) as file:
    for line in file:
      row = json.loads(line)
      records.append(row)
  return pd.DataFrame(records)


def rdf_by_dir(dir_path, exts=None, ignore_index=True, recursive=False,
               columns: (list, str) = None, info=False) -> pd.DataFrame:
  """
  从文件夹中读取所有文件并拼接成一个DataFrame
  Args:
    dir_path: 文件夹路径
    exts: 要读取的文件扩展名
    ignore_index: 是否忽略索引
    recursive: 是否循环遍历更深层级的文件夹
    columns: 指定列名
    info: 是否打印基本信息（列名、行列数）
  """
  desc = dir_path if len(dir_path) <= 23 else f'...{dir_path[-20:]}'
  path_list = []
  for p in dir_iter(dir_path, exts=exts, ignore_hidden_files=True,
                    recursive=recursive):
    path_list.append(p)
  dfs = []
  for filename in tqdm(path_list, desc=desc):
    dfs.append(rdf(filename, columns=columns))
  df = pd.concat(dfs, ignore_index=ignore_index)
  if info:
    print(f'shape: {df.shape}')
    print(f'columns: {df.columns.tolist()}')
  return df


def kml_df_create_level(gdf_dict) -> dict:
  """
  读取 dict {level: gpd.GeoDataFrame} 字典, 并根据[name]列包含关系提取
  分辨文件夹层级
  Args:
    gdf_dict: {dict} {level: gpd.GeoDataFrame} key为文件夹名称, value为对应数据
  Returns: dict: 文件夹层级字典
      example:
        {'板块名称': 0,
        'T顶豪': 1,
        'TOP1城区顶级': 2,
        'O远郊': 1,
        'O1远郊品质': 2,
        'O2远郊安居': 2}
  """
  level_dict = {}
  for i in gdf_dict:
    level_dict[i] = 0
  for k in range(1, len(gdf_dict)):
    for j in list(gdf_dict.keys())[k:]:
      sample = list(gdf_dict.keys())[k - 1]
      if set(gdf_dict[j]['name']) <= set(gdf_dict[sample]['name']):
        level_dict[j] = level_dict[sample] + 1
  return level_dict


def get_kml_df_with_level(gdf_dict) -> gpd.GeoDataFrame:
  """
  从中gdf_dict读取数据以及文件夹名称,并构造文件夹层级,按照指定层级合并所
  有数据,并打上层级标签
  Args:
    gdf_dict{dict}:{level: gpd.GeoDataFrame} key为文件夹名称, value为对应数据
  """
  level_dict = kml_df_create_level(gdf_dict)
  level_dict_new = {}
  for k, v in level_dict.items():
    columns_name = f'level_{v}'
    if columns_name not in level_dict_new:
      level_dict_new[columns_name] = []
    gdf_dict[k][columns_name] = k
    level_dict_new[columns_name].append(gdf_dict[k])
  data_all = gpd.GeoDataFrame(columns=['name', 'geometry'])
  for v in level_dict_new.values():
    level_df = pd.concat(v)
    data_all = data_all.merge(level_df, on=['name', 'geometry'], how='outer')
  data_all['name'] = data_all['name'].replace('-', None)
  return data_all


def read_kml(file_path, keep_level=True) -> gpd.GeoDataFrame:
  """
  读取kml类文件,将其转换成DataFrame, 根据separate_folders选择是否拆分,
  输出仅保留[name, geometry]字段,暂时只支持polygon和multipolygon数据
  Args:
    file_path: kml or ovkml文件路径
    keep_level: 是否保留文件夹层级标签
  """
  import kml2geojson as k2g
  features = k2g.convert(file_path, separate_folders=keep_level)
  gdf_dict = {}
  if keep_level:
    for i in range(len(features)):
      gdf = gpd.GeoDataFrame.from_features(features[i]['features'])
      level = features[i]['name']
      if 'name' not in gdf.columns:
        gdf['name'] = '-'
      gdf = gdf[['name', 'geometry']]
      gdf_dict[level] = gdf
    data_all = get_kml_df_with_level(gdf_dict)
  else:
    data_all = gpd.GeoDataFrame.from_features(features[0]['features'])
    data_all = data_all[['name', 'geometry']]
  return data_all


def read_shapefile_with_driver(file_path, encoding='utf-8') -> gpd.GeoDataFrame:
  """读取shapefile文件，分别读取dbf中的属性表和shp中的geometry进行拼接"""
  from osgeo import ogr
  from simpledbf import Dbf5

  filename = path_name(file_path)
  # 读取.dbf文件
  df = Dbf5(f'{filename}.dbf', codec=encoding).to_dataframe()
  df = gpd.GeoDataFrame(df)
  # 读取并转换shp文件中的geometry
  geoms = []
  driver = ogr.GetDriverByName('ESRI Shapefile')
  data_source = driver.Open(f'{filename}.shp', 0)  # 0表示只读模式
  layer = data_source.GetLayer()
  for i, feature in enumerate(layer):
    geom_wkt = feature.GetGeometryRef().ExportToWkt()
    geoms.append(wkt_loads(geom_wkt))
  df['geometry'] = geoms
  return gpd.GeoDataFrame(df)


def read_shapefile(file_path, **kwargs):
  """读取shapefile或geojson文件"""
  encoding = kwargs.get('encoding', 'utf-8')
  nrows = kwargs.get('nrows', None)
  try:
    df = gpd.GeoDataFrame.from_file(file_path, encoding=encoding, rows=nrows)
  except ValueError as e:
    if extension(file_path) != '.geojson':
      logging.warning(f'常规读取方式读取失败，尝试其他方式读取，{e}')
      df = read_shapefile_with_driver(file_path, encoding=encoding)
    else:
      raise ValueError(f'文件读取失败，{e}')
  finally:
    logging.warning(f'请检查输出结果并指定正确的encoding，当前为"{encoding}"')
  return df



import os
import re

from setuptools import find_packages
from setuptools import setup

pwd = os.path.dirname(__file__)

with open(os.path.join(pwd, 'src', 'ricco', '__init__.py')) as f:
  VERSION = (
    re.compile(r""".*__version__ = ["'](.*?)['"]""", re.S)
    .match(f.read())
    .group(1)
  )

with open(os.path.join(pwd, 'README.md'), encoding='utf-8') as f:
  README = f.read()

setup(
    name='ricco',
    version=VERSION,
    description='A handy ETL&GEOM kit',
    long_description=README,
    long_description_content_type="text/markdown",
    author="Ricco Wang",
    author_email="wyk_0610@163.com",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    platforms='any',
    install_requires=[
      'geojson',
      'numpy>=0.17',
      'pandas>=1.0.0',
      'requests',
      'tqdm>=4.62.0',
    ],
    classifiers=[
      'Development Status :: 3 - Alpha',
      'Intended Audience :: Developers',
      'Natural Language :: English',
      'Operating System :: OS Independent',
      'Programming Language :: Python',
      'Programming Language :: Python :: 3',
      'Topic :: Software Development :: Libraries',
    ],
    url='https://github.com/Ricco1010/ricco',
)



import builtins
import datetime
import json
import logging
import random
import re
import uuid
import warnings
from ast import literal_eval
from itertools import chain

import numpy as np
import pandas as pd

from .base import ensure_list
from .base import is_empty
from .base import not_empty
from .decorator import check_null


def to_json_string(string, errors='raise'):
  """将字符串转为接送格式的字符串"""
  assert errors in ('raise', 'coerce', 'ignore'), '参数错误'
  if not isinstance(string, (list, dict)):
    try:
      string = literal_eval(string)
    except ValueError as e:
      if errors == 'raise':
        raise ValueError(e)
      if errors == 'coerce':
        return
      if errors == 'ignore':
        return string

  return json.dumps(string, ensure_ascii=False)


def relstrip(string, kwd):
  """通过正则变大是删除左侧字符串"""
  return re.sub(f'^{kwd}', '', string)


def rerstrip(string, kwd):
  """通过正则变大是删除右侧字符串"""
  return re.sub(f'{kwd}$', '', string)


def get_shortest_element(elements: list):
  """获取列表中长度最短的元素"""

  @check_null(default_rv=np.inf)
  def condition(string):
    string = str(string)
    return len(string)

  return min(elements, key=condition)


def and_(*conditions):
  if len(conditions) < 1:
    raise Exception('最少一个条件')
  res = conditions[0]
  for c in conditions[1:]:
    res = res & c
  return res


def or_(*conditions):
  cond = []
  for c in conditions:
    if isinstance(c, (list, tuple)):
      cond.extend(list(c))
    else:
      cond.append(c)
  if len(cond) < 1:
    raise Exception('最少一个条件')
  res = cond[0]
  for c in cond[1:]:
    res = res | c
  return res


def physical_age(birthday: datetime.datetime,
                 deadline: datetime.datetime = None):
  """计算周岁，默认按照当前时间点计算"""
  now = deadline if deadline else datetime.datetime.now()
  now_str = now.strftime('%m%d')
  birth_str = birthday.strftime('%m%d')
  if now_str >= birth_str:
    return now.year - birthday.year
  else:
    return now.year - birthday.year - 1


def first_notnull_value(series):
  """筛选第一个不为空的值"""
  for v in series:
    if not_empty(v):
      return v
  warnings.warn('所有值均为空值')


@check_null()
def pinyin(word: str) -> str:
  """将中文转换为汉语拼音"""
  import pypinyin
  assert isinstance(word, str), '输入参数必须为字符串'
  s = ''
  for i in pypinyin.pinyin(word, style=pypinyin.NORMAL):
    s += ''.join(i)
  return s


@check_null(default_rv=False)
def is_valid_uuid(uuid_to_test, version=4):
  """判断一个字符是否是uuid"""
  try:
    uuid_obj = uuid.UUID(uuid_to_test, version=version)
  except ValueError:
    return False
  return str(uuid_obj) == uuid_to_test


def get_uuid(s):
  """针对格式错误的uuid和空白值生成新的uuid"""
  if is_empty(s) or not is_valid_uuid(s):
    return uuid.uuid4()
  return s


def per2float(string: str) -> float:
  """带有百分号的数值字符串转小数点形式的数值，没有百分号的返回原值"""
  if '%' in string:
    string = string.rstrip('%')
    return float(string) / 100
  return float(string)


@check_null()
def extract_num(string: str,
                num_type: str = 'str',
                method: str = 'list',
                join_list: bool = False,
                ignore_pct: bool = True,
                multi_warning=False):
  """
  提取字符串中的数值，默认返回所有数字组成的列表
  Args:
    string: 输入的字符串
    num_type:  输出的数字类型，int/float/str，默认为str
    method: 结果计算方法，对结果列表求最大/最小/平均/和等，numpy方法，默认返回列表本身
    join_list: 是否合并列表，默认FALSE
    ignore_pct: 是否忽略百分号，默认True
    multi_warning: 当有多个值的时候是否输出警告信息
  """
  assert num_type in ('float', 'int', 'str')
  string = str(string)
  ls = re.findall(r'\d+\.?\d*' if ignore_pct else r'\d+\.?\d*%?', string)
  if not ls:
    return
  if method == 'list':
    return ls if num_type == 'str' else [
      getattr(builtins, num_type)(per2float(i)) for i in ls
    ]
  if multi_warning and len(ls) >= 2:
    warnings.warn(f'有多个值进行了"{method}"运算')
  ls = [float(per2float(i)) for i in ls]
  return getattr(np, method)(ls)


@check_null()
def to_float(string,
             rex_method: str = 'mean',
             ignore_pct: bool = False,
             multi_warning=True):
  """字符串转换为float"""
  return extract_num(string,
                     num_type='float',
                     method=rex_method,
                     ignore_pct=ignore_pct,
                     multi_warning=multi_warning)


@check_null()
def house_type_format(x):
  """
  通过正则表达式将户型统一修改为1房，2房···5房及以上，目前只支持9室以下户型，
  其中5室及以上的类别为“5房及以上”
  """
  from ..resource import UTIL_CN_NUM
  if re.match('^[1-4][室房]$', x) or x == '5房及以上':
    return x
  exp = ''.join(UTIL_CN_NUM.keys())
  if res_num := re_fast(fr'([{exp}\d+])[室房]', str(x)):
    for ori, dst in UTIL_CN_NUM.items():
      res_num = res_num.replace(ori, dst)
    if int(res_num) <= 4:
      return f'{res_num}房'
    else:
      return '5房及以上'


@check_null()
def segment(x: (int, float),
            gap: (list, tuple, set, float, int),
            sep: str = '-',
            unit: str = '',
            bottom: str = '以下',
            top: str = '以上') -> str:
  """
  区间段划分工具
  Args:
    x: 数值
    gap: 间隔，固定间隔或列表
    unit: 单位，末尾
    sep: 分隔符，中间
    bottom: 默认为“以下”：80米以下
    top: 默认为“以上”：100米以上
  Returns: 区间段 'num1分隔符num2单位'：‘80-100米’
  """

  def between_list(_x, lis):
    for i in reversed(range(len(lis) - 1)):
      if _x >= lis[i]:
        return lis[i], lis[i + 1]

  assert isinstance(gap, (list, tuple, set, int, float))

  x = to_float(x)
  if isinstance(gap, (list, tuple, set)):
    gap = sorted(list(set(gap)))
    if x < gap[0]:
      return f'{gap[0]}{unit}{bottom}'
    if x >= gap[-1]:
      return f'{gap[-1]}{unit}{top}'
    return f'{between_list(x, gap)[0]}{sep}{between_list(x, gap)[1]}{unit}'

  if isinstance(gap, (int, float)):
    if x >= 0:
      return f'{int(x / gap) * gap}{sep}{int(x / gap) * gap + gap}{unit}'
    else:
      return f'{int(x / gap) * gap - gap}{sep}{int(x / gap) * gap}{unit}'


def to_str_list(series: (list, pd.Series, tuple)):
  """将列表中的元素保留为字符串、唯一、非空"""
  return list(set([str(i) for i in series if not_empty(i)]))


@check_null(default_rv=(None, None, None, None))
def fuzz_match(string: str,
               string_set: (list, pd.Series, tuple),
               fix_string_set: bool = False,
               valid_score: int = 0):
  """
  为某一字符串从某一集合中匹配相似度最高的元素
  Args:
    string: 输入的字符串
    string_set: 要去匹配的集合
    fix_string_set: 是否修复string_set中的异常数据，使用该选项会降低性能
    valid_score: 相似度大于该值的才返回
  Returns: 字符串及相似度组成的列表
  """
  from fuzzywuzzy import fuzz
  w = 0.3

  def score(x, string):
    return fuzz.partial_ratio(x, string) + fuzz.ratio(x, string) * w

  if fix_string_set:
    string_set = to_str_list(string_set)
  if is_empty(string_set):
    return None, None, None, None
  if string in string_set:
    return string, 100, 100, 100.0
  max_s = max(string_set, key=lambda x: score(x, string))

  ratio = fuzz.ratio(string, max_s)
  p_ratio = fuzz.partial_ratio(string, max_s)
  if ratio < valid_score and p_ratio < valid_score:
    return None, None, None, None
  return max_s, ratio, p_ratio, round((ratio * w + p_ratio) / (1 + w), 2)


def get_city_id_by_name(city: str):
  """获取城市代码"""
  from ..resource.city_id import CITY_IDS
  city_id = CITY_IDS.get(city if city.endswith('市') else f'{city}市')
  if not city_id:
    logging.warning('获取city_id失败，请补充该城市')
  return city_id


def sort_by_list(src_list, by_list, filter_=False) -> list:
  """
  根据一个列表对另一个列表进行筛选或排序，参照列表中不存在的元素按照原始顺序排列在后
  Args:
    src_list: 要进行排序的列表
    by_list: 参照的列表
    filter_: 是否根据参照列表筛选
  """
  res = [i for i in by_list if i in src_list]
  if not filter_:
    _res = [i for i in src_list if i not in res]
    res.extend(_res)
  return res


def remove_null_in_dict(dic: dict) -> dict:
  """将字典中值为空的元素删掉"""
  return {
    k: v
    for k, v in dic.items()
    if not_empty(v)
  }


def union_str_v2(*strings, sep='') -> str:
  """
  连接字符串，空白字符串会被忽略
  Examples:
    >>> union_str_v2('a', 'b', sep='-') # 'a-b'
    >>> union_str_v2('a', 'b', 'c') # 'abc'
  """
  if strings := [i for i in strings if not_empty(i) and i != '']:
    return sep.join(strings)


def union_list_v2(*lists) -> list:
  """合并列表"""
  lists = [ensure_list(i) for i in lists]
  return list(chain(*lists))


@check_null()
def eval_(x: str):
  """将文本型的列表、字典等转为真正的类型"""
  return literal_eval(x)


@check_null(default_rv={})
def list2dict(x: list):
  """列表转为字典，key为元素的顺序"""
  return {i: j for i, j in enumerate(x)}


def re_fast(pattern, string, warning=True):
  """根据正则表达式快速提取匹配到的第一个"""
  if isinstance(string, str):
    if ls := re.findall(pattern, string):
      if warning and len(ls) >= 2:
        warnings.warn('匹配到多个值，默认返回第一个')
      return ls[0]


def random_name():
  """随机生成中文名字，仅生成2或3字名字"""
  from ..resource.names import FirstName
  from ..resource.names import LastName
  c = [random.choice(FirstName)]
  for i in range(random.randint(1, 2)):
    c.append(random.choice(LastName))
  return ''.join(c)


def random_date():
  """获取随机日期"""
  now = datetime.datetime.now()
  tsp = int(datetime.datetime.timestamp(now))
  tsp = random.randrange(-946800000, tsp)
  return datetime.datetime.fromtimestamp(tsp).strftime("%Y%m%d")


def random_room_number(unit=False):
  """生成随机的房间号"""
  floor = random.randint(1, 17)
  room = random.randint(1, 4)
  room = str(room).zfill(2)
  if unit:
    _u = random.randint(1, 6)
    _unit = f'{_u}单元'
    return f'{_unit}{floor}{room}'
  return f'{floor}{room}'


def random_by_prob(mapping: dict):
  """根据概率生成随机值"""
  return np.random.choice(
      list(mapping.keys()),
      p=np.array(list(mapping.values())).ravel()
  )


@check_null()
def rstrip_d0(x):
  """删除末尾的‘.0’,并转为str格式，适用于对手机号等场景，如：'130.0' -> '130'"""
  _x = str(x)
  if re.match(r'^\d+\.0$', _x):
    return _x[:-2]
  return x


@check_null()
def fix_str(x: str) -> (str, None):
  """将字符串两端的空格及换行符删除，如果为空白字符串则返回空值"""
  if isinstance(x, str):
    x = x.strip()
    if x == '':
      return
  return x


def interchange_dict(dic: dict) -> dict:
  """将字典的key和value互换"""
  return dict((v, k) for k, v in dic.items())


def to_bool(x,
            na: bool = False,
            other: (bool, str) = False,
            t_list: list = None,
            f_list: list = None):
  """
  将常见的布尔类型的代替值转为布尔类型
  Args:
    x: 输入值
    na: 空值返回真还是假
    other: 无法判断的值如何处理
      - 'raise'：抛出异常
      - 'coerce'：返回传入的值
      - 除'raise'和'coerce'之外的其他值：直接返回该值
    t_list:指定为True的类别
    f_list:指定为False的类别
  """
  if not t_list:
    t_list = ['是', 1, 1.0, '1', '1.0', 't', 'true']
  if not f_list:
    f_list = ['否', 0, '0', 'f', 'false']

  if is_empty(x):
    return na
  if isinstance(x, bool):
    return x

  x2 = x.lower() if isinstance(x, str) else x
  if x2 in t_list:
    return True
  if x2 in f_list:
    return False
  if other == 'raise':
    raise ValueError(f'无法转换的值：{x}')
  if other == 'coerce':
    return x
  return other


def is_unique_series(df: pd.DataFrame,
                     key_cols: (str, list) = None,
                     ignore_na=False):
  """判断是否唯一"""
  df = df.copy()
  if not key_cols:
    key_cols = df.columns.tolist()
  key_cols = ensure_list(key_cols)
  if ignore_na:
    df = df[and_(*[df[c].notna() for c in key_cols])]
  return not df.duplicated(subset=key_cols, keep=False).any()


@check_null(default_rv=False)
def is_digit(x) -> bool:
  """判断一个值是否能通过float方法转为数值型"""
  if isinstance(x, (float, int)):
    return True
  if isinstance(x, str):
    if re.match(r'\d+\.?\d*', x):
      return True
  return False


@check_null(default_rv=False)
def is_hex(string) -> bool:
  """判断一个字符串是否是十六进制"""
  if re.match('^[0-9a-fA-F]+$', str(string)):
    return True
  return False


def isinstance_in_list(value: list, types: (str, list)):
  """检查列表内的元素类型"""
  assert isinstance(value, (list, tuple))
  return all([isinstance(v, types) for v in value])
